"""WhiskiWrap provides tools for efficiently running whisk.

This module contains the following sub-modules:
    base - The basic functions for interacting with whisk. Everything is
        imported from base into the main WhiskiWrap namespace.
    tests - Benchmarks for running whiski
    utils - utility functions for dealing with files and programs on the
        system
    video_utils - functions for dealing with video files, usually via
        system calls to ffmpeg

To read Photonfocus double-rate files, you need to install libpfdoublerate
This requires libboost_thread 1.50 to be installed to /usr/local/lib
And the libpfDoubleRate.so in this module's directory

Here is an example workflow:

import WhiskiWrap
import tables

# Run the benchmarks
test_results, durations = WhiskiWrap.tests.run_standard_benchmarks()

# Run on data
WhiskiWrap.pipeline_trace(
    input_video_filename='my_input_video.mp4',
    output_hdf5_filename='traced_whiskers.hdf5',
    chunk_sz_frames=300, epoch_sz_frames=3000, frame_start=0, frame_stop=6000,
    n_trace_processes=4)

# Load the results of that analysis
with tables.open_file('traced_whiskers.hdf5') as fi:
    test_results = pandas.DataFrame.from_records(fi.root.summary.read()) 
"""

from . import base
from . import tests
#import video_utils
from . import utils
import importlib
importlib.reload(base)
from .base import *


"""Utility functions for processing video

process_chunks_of_video : used in this module to load an input video with
    ffmpeg and dump tiff stacks to disk of each chunk.
"""
import os
import numpy as np
import subprocess
import pandas
import re
import datetime
from wwutils.video import get_video_params

def process_chunks_of_video(filename,
    frame_start=None, frame_stop=None, n_frames=None, frame_rate=None,
    frame_func=None, chunk_func=None,
    image_w=None, image_h=None,
    verbose=False,
    frames_per_chunk=1000, bufsize=-1, #10**9,
    pix_fmt='gray', finalize='list'):
    """Read frames from video, apply function, return result

    This has some advantage over wwutils.video.process_chunks_of_video
    but is out of date and not really being used.

    The dataflow is:
    1) Use a pipe to ffmpeg to load chunks of frames_per_chunk frames
    2) Apply frame_func to each frame
    3) Apply chunk_func to the chunk
    4) Append the result of chunk_func to a list
    5) "Finalize" that list and return

    If n_frames > # available, returns just the available frames with a
    warning.

    filename : file to read
    frame_start, frame_stop, n_frames : frame range to process
        If frame_start is None: defaults to zero
        If frame_stop is None: defaults to frame_start + n_frames
        If frame_stop and n_frames are both None: processes the entire video
    frame_rate : used to convert frame_start etc. to times, as required by
        ffmpeg. If None, it will be inferred from ffprobe
    frame_func : function to apply to each frame
        If None, nothing is applied. This obviously requires a lot of memory.
    chunk_func : function to apply to each chunk
        If None, nothing is applied
    image_w, image_h : width and height of video in pixels
        If None, these are inferred using ffprobe
    verbose : If True, prints out frame number for every chunk
    frames_per_chunk : number of frames to load at once from ffmpeg
    bufsize : sent to subprocess.Popen
    pix_fmt : Sent to ffmpeg.
    finalize : Function applied to final result.
        'concatenate' : apply np.concatenate
        'list' : do nothing. In this case the result will be a list of
            length n_chunks, each element of which is an array of length
            frames_per_chunk
        'listcomp' : uses a list comprehension to collapse over the chunks,
            so the result is a list of length equal to the total number of
            frames processed

    This function has been modified from wwutils.video to be optimized for
    processing chunks rather than entire videos.

    Returns: result, as described above
    """
    # Get aspect
    if image_w is None or image_h is None:
        image_w, image_h, junk = get_video_params(filename)
    if frame_rate is None:
        frame_rate = get_video_params(filename)[2]
    print('get_video_params(filename) = ', get_video_params(filename))
    # Frame range defaults
    if frame_start is None:
        frame_start = 0
    if frame_stop is None:
        if n_frames is None:
            frame_stop = np.inf
            n_frames = np.inf
        else:
            frame_stop = n_frames - frame_start
    if n_frames is None:
        n_frames = frame_stop - frame_start
    assert n_frames == frame_stop - frame_start
    if frame_stop < frame_start:
        raise ValueError("frame start cannot be greater than frame stop")

    # Set up pix_fmt
    if pix_fmt == 'gray':
        bytes_per_pixel = 1
        reshape_size = (image_h, image_w)
    elif pix_fmt == 'rgb24':
        bytes_per_pixel = 3
        reshape_size = (image_h, image_w, 3)
    else:
        raise ValueError("can't handle pix_fmt:", pix_fmt)
    print('pix_fmt = ', pix_fmt)
    read_size_per_frame = bytes_per_pixel * image_w * image_h

    # ffmpeg requires start time and total time to be in seconds, not frames
    # It seems to round up .. if I want start_time to be 0.0 and I set it to
    # 0.003 (1/10th of a frame), then the first frame is skipped.
    start_frame_time = (frame_start - 0.1) / float(frame_rate)
    total_time = (n_frames + 0.2) / float(frame_rate)

    # Create the command
    command = ['ffmpeg',
        '-ss', '%0.4f' % start_frame_time,
        '-i', filename,
        '-t', '%0.4f' % total_time,
        '-f', 'image2pipe',
        '-pix_fmt', pix_fmt,
        '-vcodec', 'rawvideo', '-']

    # To store result
    res_l = []
    frames_read = 0

    # Init the pipe
    # We set stderr to PIPE to keep it from writing to screen
    # Do this outside the try, because errors here won't init the pipe anyway
    print('comand = ', command)
    pipe = subprocess.Popen(command,
        stdout=subprocess.PIPE, stderr=subprocess.PIPE,
        bufsize=bufsize)

    # Catch any IO errors and restore stdout
    try:
        # Read in chunks
        out_of_frames = False
        print('----> n_frames: ', n_frames)

        while frames_read < n_frames and not out_of_frames:
            if verbose:
                print(frames_read)
            # Figure out how much to acquire
            if frames_read + frames_per_chunk > n_frames:
                this_chunk = n_frames - frames_read
            else:
                this_chunk = frames_per_chunk

            # Read this_chunk, or as much as we can
            raw_image = pipe.stdout.read(read_size_per_frame * this_chunk)
            # check if we ran out of frames
            if len(raw_image) < read_size_per_frame * this_chunk:
                print("warning: ran out of frames")
                out_of_frames = True
                this_chunk = len(raw_image) // read_size_per_frame
                assert this_chunk * read_size_per_frame == len(raw_image)

            # Process
            flattened_im = np.fromstring(raw_image, dtype='uint8')
            if bytes_per_pixel == 1:
                print(this_chunk, image_h, image_w)
                video = flattened_im.reshape(
                    (this_chunk, image_h, image_w))
            else:
                video = flattened_im.reshape(
                    (this_chunk, image_h, image_w, bytes_per_pixel))

            # Apply the frame_func to each frame
            # We make it an array again, but note this can lead to
            # dtype and shape problems later for some frame_func
            if frame_func is not None:
                chunk_res = np.asarray(list(map(frame_func, video)))
            else:
                chunk_res = video

            # Apply chunk_func to each chunk
            if chunk_func is not None:
                chunk_res2 = chunk_func(chunk_res)
            else:
                chunk_res2 = chunk_res

            # Store the result
            res_l.append(chunk_res2)

            # Update
            frames_read += this_chunk

    except:
        raise

    finally:
        # Restore stdout
        pipe.terminate()

        # Keep the leftover data and the error signal (ffmpeg output)
        stdout, stderr = pipe.communicate()

    if frames_read != n_frames:
        # This usually happens when there's some rounding error in the frame
        # times
        print(frames_read ,"!=", n_frames)
        raise ValueError("did not read the correct number of frames")

    # Stick chunks together
    if len(res_l) == 0:
        print("warning: no data found")
        res = np.array([])
    elif finalize == 'concatenate':
        res = np.concatenate(res_l)
    elif finalize == 'listcomp':
        res = np.array([item for sublist in res_l for item in sublist])
    elif finalize == 'list':
        res = res_l
    else:
        print("warning: unknown finalize %r" % finalize)
        res = res_l

    return res



"""Generating or processing video, often using ffmpeg"""
from __future__ import print_function
from __future__ import division
from builtins import zip
from builtins import str
from builtins import map
from builtins import input
from builtins import object
from past.utils import old_div
import numpy as np
import subprocess
import re
import datetime
import os
import ffmpeg
import math
from time import gmtime

class OutOfFrames(BaseException):
    """Exception raised when more frames cannot be extracted from a video"""
    pass

def ffmpeg_frame_string(filename, frame_time=None, frame_number=None):
    """Given a frame time or number, create a string for ffmpeg -ss.
    
    This attempts to reverse engineer the way that ffmpeg converts frame
    times to frame numbers, so that we can specify an exact frame number
    and get that exact frame.
    
    As far as I can tell, if you request time T, 
    ffmpeg rounds T to the nearest millisecond, 
    and then gives you frame N, 
    where N is ceil(T * frame_rate).
    
    So -.001 gives you the first frame, and .001 gives you the second frame.
    
    It's hard to predict what will happen within one millisecond of a
    frame time, so try to avoid that if exactness is important.
    
    
    filename : video file. Used to get frame rate.
    
    frame_time : This one takes precedence if both are provided.
        We simply subtract half of the frame interval, and then round to
        the nearest millisecond to account for ffmpeg's rounding up.
    
    frame_number : This one is used if frame_time is not specified.
        We convert to a frame time using
            ((frame_number / frame_rate) - 1 ms) 
            rounded down to the nearest millisecond.
        This should give accurate results as long as frame rate is not
        >500 fps or so.
    
    frametime, frame_number : which frame to get
        if you request time T, ffmpeg gives you frame N, where N is 
        ceil(time * frame_rate). So -.001 gives you the first frame, and
        .001 gives you the second frame. It's hard to predict what will
        happen with one ms of the exact frame time due to rounding errors.
    
    Returns : string, suitable for -ss
    """
    if frame_number is not None:
        # If specified by number, convert to time
        frame_rate = get_video_params(filename)[2]
        use_frame_time = (old_div(frame_number, float(frame_rate))) - .001
        use_frame_time = old_div(np.floor(use_frame_time * 1000), 1000.)
    
    elif frame_time is not None:
        frame_rate = get_video_params(filename)[2]
        use_frame_time = frame_time - (old_div(1., (2 * frame_rate)))
    
    else:
        raise ValueError("must specify frame by time or number")
    
    use_frame_string = '%0.3f' % use_frame_time
    return use_frame_string

def get_frame(filename, frametime=None, frame_number=None, frame_string=None,
    pix_fmt='gray', bufsize=10**9, path_to_ffmpeg='ffmpeg', vsync='drop'):
    """Returns a single frame from a video as an array.
    
    This creates an ffmpeg process and extracts data from it with a pipe.

    filename : video filename
    frame_string : to pass to -ss
    frametime, frame_number:
        If frame_string is None, then these are passed to 
        ffmpeg_frame_string to generate a frame string.
        
    pix_fmt : the "output" format of ffmpeg.
        currently only gray and rgb24 are accepted, because I need to 
        know how to reshape the result.
    
    This syntax is used to seek with ffmpeg:
        ffmpeg -ss %frametime% -i %filename% -vframes 1 ...
    This is supposed to be relatively fast while still accurate.
    
    TODO: Get this to return multiple frames from the same instance
    
    Returns:
        frame, stdout, stderr
        frame : 2d array, of shape (height, width)
        stdout : typically blank
        stderr : ffmpeg's text output
    """
    v_width, v_height = get_video_aspect(filename)
    
    if pix_fmt == 'gray':
        bytes_per_pixel = 1
        reshape_size = (v_height, v_width)
    elif pix_fmt == 'rgb24':
        bytes_per_pixel = 3
        reshape_size = (v_height, v_width, 3)
    else:
        raise ValueError("can't handle pix_fmt:", pix_fmt)
    
    # Generate a frame string if we need it
    if frame_string is None:
        frame_string = ffmpeg_frame_string(filename, 
            frame_time=frametime, frame_number=frame_number)
    
    # Create the command
    command = [path_to_ffmpeg, 
        '-ss', frame_string,
        '-i', filename,
        '-vsync', vsync,
        '-vframes', '1',       
        '-f', 'image2pipe',
        '-pix_fmt', pix_fmt,
        '-vcodec', 'rawvideo', '-']
    
    # To store result
    res_l = []
    frames_read = 0

    # Init the pipe
    # We set stderr to PIPE to keep it from writing to screen
    # Do this outside the try, because errors here won't init the pipe anyway
    pipe = subprocess.Popen(command, 
        stdout=subprocess.PIPE, stderr=subprocess.PIPE, 
        bufsize=bufsize)

    try:
        read_size = bytes_per_pixel * v_width * v_height
        raw_image = pipe.stdout.read(read_size)    
        if len(raw_image) < read_size:
            raise OutOfFrames        
        flattened_im = np.fromstring(raw_image, dtype='uint8')
        frame = flattened_im.reshape(reshape_size)    
    
    except OutOfFrames:
        print("warning: cannot get frame")
        frame = None
    
    finally:
        # Restore stdout
        pipe.terminate()

        # Keep the leftover data and the error signal (ffmpeg output)
        stdout, stderr = pipe.communicate()    
        
        # Convert to string
        if stdout is not None:
            stdout = stdout.decode('utf-8')
        if stderr is not None:
            stderr = stderr.decode('utf-8')
    
    return frame, stdout, stderr


def frame_dump(filename, frametime, output_filename='out.png', 
    meth='ffmpeg fast', subseek_cushion=20., verbose=False, dry_run=False,
    very_verbose=False):
    """Dump the frame in the specified file.
    
    Probably better to use get_frame instead.
    
    If the subprocess fails, CalledProcessError is raised.
    Special case: if seek is beyond the end of the file, nothing is done
    and no error is raised
    (because ffmpeg does not report any problem in this case).
    
    Values for meth:
        'ffmpeg best' : Seek quickly, then accurately
            ffmpeg -y -ss :coarse: -i :filename: -ss :fine: -vframes 1 \
                :output_filename:
        'ffmpeg fast' : Seek quickly
            ffmpeg -y -ss :frametime: -i :filename: -vframes 1 :output_filename:
        'ffmpeg accurate' : Seek accurately, but takes forever
            ffmpeg -y -i :filename: -ss frametime -vframes 1 :output_filename:
        'mplayer' : This takes forever and also dumps two frames, the first 
            and the desired. Not currently working but something like this:
            mplayer -nosound -benchmark -vf framestep=:framenum: \
                -frames 2 -vo png :filename:
    
    Note that output files are always overwritten without asking.
    
    With recent, non-avconv versions of ffmpeg, it appears that 'ffmpeg fast'
    is just as accurate as 'ffmpeg best', and is now the preferred method.
    
    Use scipy.misc.imread to read them back in.
    
    Source
        https://trac.ffmpeg.org/wiki/Seeking%20with%20FFmpeg
    """
    
    if meth == 'mplayer':
        raise ValueError("mplayer not supported")
    elif meth == 'ffmpeg best':
        # Break the seek into a coarse and a fine
        coarse = np.max([0, frametime - subseek_cushion])
        fine = frametime - coarse
        syscall = 'ffmpeg -y -ss %r -i %s -ss %r -vframes 1 %s' % (
            coarse, filename, fine, output_filename)
    elif meth == 'ffmpeg accurate':
        syscall = 'ffmpeg -y -i %s -ss %r -vframes 1 %s' % (
            filename, frametime, output_filename)
    elif meth == 'ffmpeg fast':
        syscall = 'ffmpeg -y -ss %r -i %s -vframes 1 %s' % (
            frametime, filename, output_filename)
    
    if verbose:
        print(syscall)
    if not dry_run:
        #os.system(syscall)
        syscall_l = syscall.split(' ')
        syscall_result = subprocess.check_output(syscall_l, 
            stderr=subprocess.STDOUT)
        if very_verbose:
            print(syscall_result)

def process_chunks_of_video(filename, n_frames, func='mean', verbose=False,
    frame_chunk_sz=1000, bufsize=10**9,
    image_w=None, image_h=None, pix_fmt='gray',
    finalize='concatenate', path_to_ffmpeg='ffmpeg', vsync='drop'):
    """Read frames from video, apply function, return result
    
    Uses a pipe to ffmpeg to load chunks of frame_chunk_sz frames, applies
    func, then stores just the result of func to save memory.
    
    If n_frames > # available, returns just the available frames with a
    warning.
    
    filename : file to read
    n_frames : number of frames to process
        if None or np.inf, will continue until video is exhausted
    func : function to apply to each frame
        If 'mean', then func = lambda frame: frame.mean()
        If 'keep', then func = lambda frame: frame
        'keep' will return every frame, which will obviously require a lot
        of memory.
    verbose : If True, prints out frame number for every chunk
    frame_chunk_sz : number of frames to load at once from ffmpeg
    bufsize : sent to subprocess.Popen
    image_w, image_h : width and height of video in pxels
    pix_fmt : Sent to ffmpeg
    """
    if n_frames is None:
        n_frames = np.inf
    
    # Default function is mean luminance
    if func == 'mean':
        func = lambda frame: frame.mean()
    elif func == 'keep':
        func = lambda frame: frame
    elif func is None:
        raise ValueError("must specify frame function")
    
    # Get aspect
    if image_w is None:
        image_w, image_h = get_video_aspect(filename)
    
    # Set up pix_fmt
    if pix_fmt == 'gray':
        bytes_per_pixel = 1
        reshape_size = (image_h, image_w)
    elif pix_fmt == 'rgb24':
        bytes_per_pixel = 3
        reshape_size = (image_h, image_w, 3)
    else:
        raise ValueError("can't handle pix_fmt:", pix_fmt)
    read_size_per_frame = bytes_per_pixel * image_w * image_h
    
    # Create the command
    command = [path_to_ffmpeg,
        '-i', filename,
        '-vsync', vsync,
        '-f', 'image2pipe',
        '-pix_fmt', pix_fmt,
        '-vcodec', 'rawvideo', '-']
    
    # To store result
    res_l = []
    frames_read = 0

    # Init the pipe
    # We set stderr to PIPE to keep it from writing to screen
    # Do this outside the try, because errors here won't init the pipe anyway
    # Actually, stderr will fill up and the process will hang
    # http://stackoverflow.com/questions/375427/non-blocking-read-on-a-subprocess-pipe-in-python/4896288#4896288
    pipe = subprocess.Popen(command, 
        stdout=subprocess.PIPE, stderr=open(os.devnull, 'w'), 
        bufsize=bufsize)

    # Catch any IO errors and restore stdout
    try:
        # Read in chunks
        out_of_frames = False
        while frames_read < n_frames and not out_of_frames:
            if verbose:
                print(frames_read)
            # Figure out how much to acquire
            if frames_read + frame_chunk_sz > n_frames:
                this_chunk = n_frames - frames_read
            else:
                this_chunk = frame_chunk_sz
            
            # Read this_chunk, or as much as we can
            raw_image = pipe.stdout.read(read_size_per_frame * this_chunk)
            
            # check if we ran out of frames
            if len(raw_image) < read_size_per_frame * this_chunk:
                print("warning: ran out of frames")
                out_of_frames = True
                this_chunk = old_div(len(raw_image), read_size_per_frame)
                assert this_chunk * read_size_per_frame == len(raw_image)
            
            # Process
            flattened_im = np.fromstring(raw_image, dtype='uint8')
            if bytes_per_pixel == 1:
                video = flattened_im.reshape(
                    (this_chunk, image_h, image_w))
            else:
                video = flattened_im.reshape(
                    (this_chunk, image_h, image_w, bytes_per_pixel))
            
            # Store as list to avoid dtype and shape problems later
            #chunk_res = np.asarray(map(func, video))
            chunk_res = list(map(func, video))
            
            # Store
            res_l.append(chunk_res)
            
            # Update
            frames_read += this_chunk

    except:
        raise

    finally:
        # Restore stdout
        pipe.terminate()

        # Keep the leftover data and the error signal (ffmpeg output)
        stdout, stderr = pipe.communicate()
        
        # Convert to string
        if stderr is not None:
            stderr = stderr.decode('utf-8')

    if not np.isinf(n_frames) and frames_read != n_frames:
        # This usually happens when there's some rounding error in the frame
        # times
        raise ValueError("did not read the correct number of frames")

    # Stick chunks together
    if len(res_l) == 0:
        print("warning: no data found")
        res = np.array([])
    elif finalize == 'concatenate':
        res = np.concatenate(res_l)
    elif finalize == 'listcomp':
        res = np.array([item for sublist in res_l for item in sublist])
    elif finalize == 'list':
        res = res_l
    else:
        print("warning: unknown finalize %r" % finalize)
        res = res_l
        
    return res

def get_video_aspect(video_filename):
    """Returns width, height of video using ffmpeg-python"""
    if not os.path.exists(video_filename):
        raise ValueError("%s does not exist" % video_filename)
    
    probe = ffmpeg.probe(video_filename)
    assert len(probe['streams']) == 1
    width = probe['streams'][0]['width']
    height = probe['streams'][0]['height']
    
    return width, height

def get_video_frame_rate(video_filename):
    """Returns frame rate of video using ffmpeg-python
    
    https://video.stackexchange.com/questions/20789/ffmpeg-default-output-frame-rate
    """
    if not os.path.exists(video_filename):
        raise ValueError("%s does not exist" % video_filename)
    
    probe = ffmpeg.probe(video_filename)
    assert len(probe['streams']) == 1
    
    # Seems to be two ways of coding, not sure which is better
    avg_frame_rate = probe['streams'][0]['avg_frame_rate']
    r_frame_rate = probe['streams'][0]['r_frame_rate']
    assert avg_frame_rate == r_frame_rate
    
    # Convert fraction to number
    num, den = avg_frame_rate.split('/')
    frame_rate = float(num) / float(den)
    
    return frame_rate

def get_video_params(video_filename, crop=None):
    """Returns width, height, frame_rate of video using ffmpeg-python"""
    # if crop is not None:
    #     width, height = crop[0], crop[1]
    # else:
    width, height = get_video_aspect(video_filename)
    frame_rate = get_video_frame_rate(video_filename)
    return width, height, frame_rate

def get_video_duration(video_filename):
    if not os.path.exists(video_filename):
        raise ValueError("%s does not exist" % video_filename)
    
    probe = ffmpeg.probe(video_filename)
    print("=================:>>  PROBE",  probe)
    assert len(probe['streams']) == 1
    
    # Container duration
    container_duration = float(probe['format']['duration'])
    
    # Stream duration
    stream_duration_s = probe['streams'][0]['duration']
    
    # For some reason this is in nanoseconds, convert to microseconds
    stream_duration_s = stream_duration_s[:-3]
    
    # Match
    try:
        video_duration_temp = datetime.datetime.strptime(
            stream_duration_s, '%H:%M:%S.%f')
        stream_duration_dt = datetime.timedelta(
            hours=video_duration_temp.hour,
            minutes=video_duration_temp.minute,
            seconds=video_duration_temp.second,
            microseconds=video_duration_temp.microsecond)

        stream_duration = stream_duration_dt.total_seconds()

    except ValueError as e:
        print('Warning: ', e)
        stream_duration = float(stream_duration_s)

    assert math.isclose(stream_duration, container_duration, rel_tol=1e-3), f"stream_duration ({stream_duration}) != container_duration ({container_duration}), relative {stream_duration/container_duration} "
    
    return stream_duration

def choose_rectangular_ROI(vfile, n_frames=4, interactive=False, check=True,
    hints=None):
    """Displays a subset of frames from video so the user can specify an ROI.
    
    If interactive is False, the frames are simply displayed in a figure.
    If interactive is True, a simple text-based UI allows the user to input
    the x- and y- coordinates of the ROI. These are drawn and the user has
    the opportunity to confirm them.
    
    If check is True, then the values are swapped as necessary such that
    x0 < x1 and y0 < y1.
    
    Finally the results are returned as a dict with keys x0, x1, y0, y1.
    
    hints : dict, or None
        If it has key x0, x1, y0, or y1, the corresponding values will
        be displayed as a hint to the user while selecting.
    """
    import matplotlib.pyplot as plt
    import wwutils.plot
    # Not sure why this doesn't work if it's lower down in the function
    if interactive:
        plt.ion() 

    # Get frames
    duration = get_video_duration(vfile)
    frametimes = np.linspace(duration * .1, duration * .9, n_frames)
    frames = []
    for frametime in frametimes:
        frame, stdout, stderr = get_frame(vfile, frametime)
        frames.append(frame)
    
    # Plot them
    f, axa = plt.subplots(1, 4, figsize=(11, 2.5))
    f.subplots_adjust(left=.05, right=.975, bottom=.05, top=.975)
    for frame, ax in zip(frames, axa.flatten()):
        wwutils.plot.imshow(frame, ax=ax, axis_call='image', cmap=plt.cm.gray)
    wwutils.plot.harmonize_clim_in_subplots(fig=f, clim=(0, 255))

    # Get interactive results
    res = {}
    if interactive:
        params_l = ['x0', 'x1', 'y0', 'y1']
        lines = []
        try:
            while True:
                for line in lines:
                    line.set_visible(False)    
                plt.draw()
                
                # Get entries for each params
                for param in params_l:
                    # Form request string, using hint if available
                    hint = None
                    if hints is not None and param in hints:
                        hint = hints[param]
                    if hint is None:
                        request_s = 'Enter %s: ' % param
                    else:
                        request_s = 'Enter %s [hint = %d]: ' % (param, hint)
                    
                    # Keep getting input till it is valid
                    while True:
                        try:
                            val = input(request_s)
                            break
                        except ValueError:
                            print("invalid entry")
                    res[param] = int(val)

                # Check ordering
                if check:
                    if res['x0'] > res['x1']:
                        res['x0'], res['x1'] = res['x1'], res['x0']
                    if res['y0'] > res['y1']:
                        res['y0'], res['y1'] = res['y1'], res['y0']

                # Draw results
                for ax in axa:
                    lines.append(ax.plot(
                        ax.get_xlim(), [res['y0'], res['y0']], 'r-')[0])
                    lines.append(ax.plot(
                        ax.get_xlim(), [res['y1'], res['y1']], 'r-')[0])
                    lines.append(ax.plot(
                        [res['x0'], res['x0']], ax.get_ylim(), 'r-')[0])            
                    lines.append(ax.plot(
                        [res['x1'], res['x1']], ax.get_ylim(), 'r-')[0])
                plt.draw()

                # Get confirmation
                choice = input("Confirm [y/n/q]: ")
                if choice == 'q':
                    res = {}
                    print("cancelled")
                    break
                elif choice == 'y':
                    break
                else:
                    pass
        except KeyboardInterrupt:
            res = {}
            print("cancelled")
        finally:
            plt.ioff()
            plt.close(f)
    
    return res    


def crop(input_file, output_file, crop_x0, crop_x1, 
    crop_y0, crop_y1, crop_stop_sec=None, vcodec='mpeg4', quality=2, 
    overwrite=True, verbose=False, very_verbose=False):
    """Crops the input file into the output file"""
    # Overwrite avoid
    if os.path.exists(output_file) and not overwrite:
        raise ValueError("%s already exists" % output_file)
    
    # Set up width, height and origin of crop zone
    if crop_x0 > crop_x1:
        crop_x0, crop_x1 = crop_x1, crop_x0
    if crop_y0 > crop_y1:
        crop_y0, crop_y1 = crop_y1, crop_y0
    width = crop_x1 - crop_x0
    height = crop_y1 - crop_y0
    
    # Form the syscall
    crop_string = '"crop=%d:%d:%d:%d"' % (width, height, crop_x0, crop_y0)
    syscall_l = ['ffmpeg', '-i', input_file, '-y',
        '-vcodec', vcodec,
        '-q', str(quality),
        '-vf', crop_string]
    if crop_stop_sec is not None:
        syscall_l += ['-t', str(crop_stop_sec)]
    syscall_l.append(output_file)

    # Call, redirecting to standard output so that we can catch it
    if verbose:
        print(' '.join(syscall_l))
    
    # I think when -t parameter is set, it raises CalledProcessError
    #~ syscall_result = subprocess.check_output(syscall_l, 
        #~ stderr=subprocess.STDOUT)
    #~ if very_verbose:
        #~ print syscall_result
    os.system(' '.join(syscall_l))

def split():
    # ffmpeg -i 150401_CR1_cropped.mp4 -f segment -vcodec copy -reset_timestamps 1 -map 0 -segment_time 1000 OUTPUT%d.mp4
    pass


class WebcamController(object):
    def __init__(self, device='/dev/video0', output_filename='/dev/null',
        width=320, height=240, framerate=30,
        window_title='webcam', image_controls=None,
        ):
        """Init a new webcam controller for a certain webcam.
        
        image_controls : dict containing controls like gain, exposure
            They will be set to reasonable defaults if not specified.
        """
        # Store params
        self.device = device
        self.output_filename = output_filename
        self.width = width
        self.height = height
        self.framerate = framerate
        self.window_title = window_title
        
        if self.output_filename is None:
            self.output_filename = '/dev/null'
        
        # Image controls
        self.image_controls = {
            'gain': 3,
            'exposure': 20,
            'brightness': 40,
            'contrast': 50,
            'saturation': 69,
            'hue': 0,
            'white_balance_automatic': 0,
            'gain_automatic': 0,
            'auto_exposure': 1, # flipped
            }
        if image_controls is not None:
            self.image_controls.update(image_controls)
        
        self.read_stderr = None
        self.ffplay_stderr = None
        self.ffplay_stdout = None
        
        self.ffplay_proc = None
        self.read_proc = None
        self.tee_proc = None
    
    def start(self, print_ffplay_proc_stderr=False, print_read_proc_stderr=False):
        """Start displaying and encoding
        
        To stop, call the stop method, or close the ffplay window.
        In the latter case, it will keep reading from the webcam until
        you call cleanup or delete the object.
        
        print_ffplay_proc_stderr : If True, prints the status messages to
            the terminal from the the process that plays video to the screen.
            If False, writes to /dev/null.
        print_read_proc_stderr : Same, but for the process that reads from
            the webcam.
        """
        # Set the image controls
        self.set_controls()
        
        # Create a process to read from the webcam
        # stdin should be pipe so it doesn't suck up keypresses (??)
        # stderr should be null, so pipe doesn't fill up and block
        # stdout will go to downstream process
        if print_read_proc_stderr:
            read_proc_stderr = None
        else:
            read_proc_stderr = open(os.devnull, 'w')
        read_proc_cmd_l = ['ffmpeg',
            '-f', 'video4linux2',
            '-i', self.device,
            '-vcodec', 'libx264',
            '-qp', '0',
            '-vf', 'format=gray',
            '-preset', 'ultrafast',
            '-f', 'rawvideo', '-',
            ] 
        self.read_proc = subprocess.Popen(read_proc_cmd_l, stdin=subprocess.PIPE, 
            stdout=subprocess.PIPE, stderr=read_proc_stderr)
        
        # Sometimes the read_proc fails because the device is busy or "Input/ouput error"
        # but the returncode isn't set or anything so I don't know how to
        # detect this.

        # Tee the compressed output to a file
        self.tee_proc = subprocess.Popen(['tee', self.output_filename], 
            stdin=self.read_proc.stdout,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        # Play the output
        if print_ffplay_proc_stderr:
            ffplay_proc_stderr = None
        else:
            ffplay_proc_stderr = open(os.devnull, 'w')        
        self.ffplay_proc = subprocess.Popen([
            'ffplay', 
            #~ '-fflags', 'nobuffer', # not compatible with analyzeduration or probesize?
            '-analyzeduration', '500000', # 500 ms delay in starting
            '-window_title', self.window_title,
            '-',
            ], 
            stdin=self.tee_proc.stdout,
            stdout=subprocess.PIPE, stderr=ffplay_proc_stderr)

        # This is supposed to allow SIGPIPE
        # https://docs.python.org/2/library/subprocess.html#replacing-shell-pipeline
        self.read_proc.stdout.close()
        self.tee_proc.stdout.close()        
    
    def set_controls(self):
        """Use v4l2-ctl to set the controls"""
        # Form the param list
        cmd_list = ['v4l2-ctl',
            '-d', self.device,
            '--set-fmt-video=width=%d,height=%d' % (self.width, self.height),
            '--set-parm=%d' % self.framerate,    
            ]
        for k, v in list(self.image_controls.items()):
            cmd_list += ['-c', '%s=%d' % (k, v)]

        # Create a process to set the parameters and run it
        self.set_proc = subprocess.Popen(cmd_list,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        self.set_stdout, self.set_stderr = self.set_proc.communicate()

        if self.set_proc.returncode != 0:
            print("failed to set parameters")
            print(self.set_stdout)
            print(self.set_stderr)
            raise IOError("failed to set parameters")
    
    def stop(self):
        if self.ffplay_proc is not None:
            self.ffplay_proc.terminate()
        self.cleanup()
    
    def update(self):
        pass
    
    def cleanup(self):
        self.__del__()
    
    def __del__(self):
        if self.ffplay_proc is not None:
            if self.ffplay_proc.returncode is None:
                self.ffplay_stdout, self.ffplay_stderr = \
                    self.ffplay_proc.communicate()
        
        if self.read_proc is not None:
            if self.read_proc.returncode is None:
                self.read_proc.terminate()
                self.read_proc.wait()
        
        if self.tee_proc is not None:
            self.tee_proc.wait()


class WebcamControllerFFplay(WebcamController):
    """Simpler version that just plays with ffplay"""
    def start(self):
        self.set_controls()
        self.ffplay_proc = subprocess.Popen([
            'ffplay',
            '-f', 'video4linux2',
            '-window_title', self.window_title,
            '-i', self.device,
            ], 
            stdout=subprocess.PIPE, stderr=open(os.devnull, 'w'),
            bufsize=1000000)
        self.stdout_l = []
        self.stderr_l = []

    def stop(self):
        self.ffplay_proc.terminate()
        self.cleanup()
    
    def update(self):
        """This is supposed to read the stuff on stderr but I can't
        get it to not block"""
        return
        #~ self.stdout_l.append(self.ffplay_proc.stdout.read())
        print("update")
        data = self.ffplay_proc.stderr.read(1000000)
        print("got data")
        print(len(data))
        while len(data) == 1000000:
            self.stderr_l.append(data)
            data = self.ffplay_proc.stderr.read(1000000)
        print("done")
    
    def __del__(self):
        try:
            if self.ffplay_proc.returncode is None:
                self.ffplay_stdout, self.ffplay_stderr = (
                    self.ffplay_proc.communicate())
        except AttributeError:
            pass


from __future__ import print_function
from __future__ import division
from builtins import str
from builtins import range
from past.utils import old_div
import numpy as np
import scipy.stats
import pandas
import statsmodels.stats.multitest

try:
    import rpy2.robjects as robjects
    r = robjects.r
except ImportError:
    # it's all good
    pass

def adjust_pval(ser, method='fdr_bh'):
    arr = statsmodels.stats.multitest.multipletests(
        ser.values, method=method)[1]
    res = pandas.Series(arr, index=ser.index)
    return res
    
def z2p(zvals):
    """Return the two-tailed p-value corresponding to a z-stat"""
    norm_rv = scipy.stats.norm()
    res = 2 * (1 - norm_rv.cdf(np.abs(zvals)))
    
    try:
        # If zvals is dataframe
        return pandas.DataFrame(res, index=zvals.index, columns=zvals.columns)
    except AttributeError:
        try:
            # If zvals is Series
            return pandas.Series(res, index=zvals.index)
        except AttributeError:
            # If zvals is anything else, like a float
            return res

def pvalue_to_significance_string(pvalue):
    """Return significance string like '*' based on pvalue"""
    if pvalue < .001:
        return '***'
    elif pvalue < .01:
        return '**'
    elif pvalue < .05:
        return '*'
    else:
        return 'n.s.'

def mad_1d(arr1d):
    """My implementation of MAD.
    
    I think there is a bug in statsmodels.robust.mad because the
    median is not subtracted from the data
    """
    deviations = arr1d - np.median(arr1d)
    return old_div(np.median(np.abs(deviations)), .6745)
    
def binom_confint(x=None, n=None, data=None, alpha=.95, meth='beta'):
    """Calculates the confidence interval for binomial data.
    
    Previously I used R's binom.confint with the 'exact' method.
    statsmodels' proportion_confint with the 'beta' method seems to be the
    same.
    """
    if data is not None:
        x = np.sum(data)
        n = len(data)
    
    if meth == 'R_exact':
        r("library(binom)")
        res = r("binom.confint(%d, %d, %f, methods='%s')" % (
            x, n, alpha, 'exact'))
        return res[res.names.index('lower')][0], res[res.names.index('upper')][0]
    else:
        import statsmodels.stats.proportion
        
        # Call statsmodels
        res = list(statsmodels.stats.proportion.proportion_confint(
            x, n, method=meth))
        
        # Bug where nan is sometimes returned
        # https://github.com/statsmodels/statsmodels/pull/2789
        if x == n:
            res[1] = 1
        if x == 0:
            res[0] = 0
        
        return tuple(res)
        

def bootstrap_regress(x, y, n_boot=1000):
    from matplotlib import mlab
    x = np.asarray(x)
    y = np.asarray(y)
    
    m_l, b_l = [], []
    for n in range(n_boot):
        msk = np.random.randint(0, len(x), size=len(x))
        m, b, rval, pval, stderr = scipy.stats.stats.linregress(x[msk], y[msk])
        m_l.append(m)
        b_l.append(b)
    
    res = {
        'slope_m': np.mean(m_l),
        'slope_l': mlab.prctile(m_l, p=2.5),
        'slope_h': mlab.prctile(m_l, p=97.5),
        'intercept_m': np.mean(b_l),
        'intercept_l': mlab.prctile(b_l, p=2.5),
        'intercept_h': mlab.prctile(b_l, p=97.5),
        }
    return res    
    


def r_adj_pval(a, meth='BH'):
    """Adjust p-values in R using specified method"""
    a = np.asarray(a)
    robjects.globalenv['unadj_p'] = robjects.FloatVector(a.flatten())
    return np.array(r("p.adjust(unadj_p, '%s')" % meth)).reshape(a.shape)

def check_float_conversion(a1, a2, tol):
    """Checks that conversion to R maintained uniqueness of arrays.
    
    a1 : array of unique values, typically originating in Python
    a2 : array of unique values, typically grabbed from R
    
    If the lengths are different, or if either contains values that
    are closer than `tol`, an error is raised.
    """
    if len(a1) != len(a2):
        raise ValueError("uniqueness violated in conversion")
    if len(a1) > 1:
        if np.min(np.diff(np.sort(a1))) < tol:
            raise ValueError("floats separated by less than tol")
        if np.min(np.diff(np.sort(a2))) < tol:
            raise ValueError("floats separated by less than tol")

def r_utest(x, y, mu=0, verbose=False, tol=1e-6, exact='FALSE', 
    fix_nan=True, fix_float=False, paired='FALSE'):
    """Mann-Whitney U-test in R
    
    This is a test on the median of the distribution of sample in x minus
    sample in y. It uses the R implementation to avoid some bugs and gotchas
    in scipy.stats.mannwhitneyu.
    
    Some care is taken when converting floats to ensure that uniqueness of
    the datapoints is conserved, which should maintain the ranking.
    
    x : dataset 1
    y : dataset 2
        If either x or y is empty, prints a warning and returns some
        values that indicate no significant difference. But note that
        the test is really not appropriate in this case.
    mu : null hypothesis on median of sample in x minus sample in y
    verbose : print a bunch of output from R
    tol : if any datapoints are closer than this, raise an error, on the
        assumption that they are only that close due to numerical
        instability
    exact : see R doc
        Defaults to FALSE since if the data contain ties and exact is TRUE,
        R will print a warning and approximate anyway
    fix_nan : if p-value is nan due to all values being equal, then
        set p-value to 1.0. But note that the test is really not appropriate
        in this case.
    fix_float : int, or False
        if False or if the data is integer, does nothing
        if int and the data is float, then the data are multiplied by this
        and then rounded to integers. The purpose is to prevent the numerical
        errors that are tested for in this function. Note that differences
        less than 1/fix_float will be removed.
    
    Returns: dict with keys ['U', 'p', 'auroc']
        U : U-statistic. 
            Large U means that x > y, small U means that y < x
            Compare scipy.stats.mannwhitneyu which always returns minimum U
        p : two-sided p-value
        auroc : area under the ROC curve, calculated as U/(n1*n2)
            Values greater than 0.5 indicate x > y
    """
    x = np.asarray(x).flatten()
    y = np.asarray(y).flatten()
    
    # What type of R object to create
    if x.dtype.kind in 'iu' and y.dtype.kind in 'iu':        
        behavior = 'integer'
    elif x.dtype.kind == 'f' or y.dtype.kind == 'f':
        behavior = 'float'
    else:
        raise ValueError("cannot determine datatype of x and y")
    
    # Optionally fix float
    if fix_float is True:
        fix_float = 1e6
    if fix_float and behavior == 'float':
        x = np.rint(x * fix_float).astype(np.int)
        y = np.rint(y * fix_float).astype(np.int)
        behavior = 'integer'
    
    # Define variables
    if behavior == 'integer':
        robjects.globalenv['x'] = robjects.IntVector(x)
        robjects.globalenv['y'] = robjects.IntVector(y)
    elif behavior == 'float':
        robjects.globalenv['x'] = robjects.FloatVector(x)
        robjects.globalenv['y'] = robjects.FloatVector(y)
    
        # Check that uniqueness is maintained
        ux_r, ux_p = r("unique(x)"), np.unique(x)
        check_float_conversion(ux_r, ux_p, tol)
        uy_r, uy_p = r("unique(y)"), np.unique(y)
        check_float_conversion(uy_r, uy_p, tol)
        
        # and of the concatenated
        uxy_r, uxy_p = r("unique(c(x,y))"), np.unique(np.concatenate([x,y]))
        check_float_conversion(uxy_r, uxy_p, tol)
    
    # Run the test
    if len(x) == 0 or len(y) == 0:
        print("warning empty data in utest, returning p = 1.0")
        U, p, auroc = 0.0, 1.0, 0.5
    else:
        res = r("wilcox.test(x, y, mu=%r, exact=%s, paired=%s)" % (mu, exact, paired))
        U, p = res[0][0], res[2][0]
        auroc = old_div(float(U), (len(x) * len(y)))
    
    # Fix p-value
    if fix_nan and np.isnan(p):
        p = 1.0
    
    # debug
    if verbose:
        print(behavior)
        s_x = str(robjects.globalenv['x'])
        print(s_x[:1000] + '...')
        s_y = str(robjects.globalenv['y'])
        print(s_y[:1000] + '...')
        print(res)
    
    return {'U': U, 'p': p, 'auroc': auroc}

def anova(df, fmla, typ=3):
    from statsmodels.formula.api import ols
    from statsmodels.stats.anova import anova_lm
    
    # Anova/OLS
    lm = ols(fmla, df).fit() # 'data' <==> 'df' keyword change with version
    
    # Grab the pvalues (note we use Type III)
    aov = anova_lm(lm, typ=typ)
    pvals = aov["PR(>F)"]
    pvals.index = ['p_' + s for s in pvals.index]
    
    # Grab the explainable sum of squares
    ess = aov.drop("Residual").sum_sq
    ess = old_div(ess, ess.sum())
    ess.index = ['ess_' + s for s in ess.index]
    
    # Grab the fit
    fit = lm.params
    fit.index = ['fit_' + s for s in fit.index]   

    # I think this happens with pathological inputs
    if np.any(aov['sum_sq'] < 0):
        old_div(1,0)

    return {'lm':lm, 'aov':aov, 'pvals':pvals, 'ess':ess, 'fit':fit}
    



from setuptools import setup

setup(
   name='WhiskiWrap',
   version='1.1.8',
   author='Chris Rodgers',
   author_email='',
   maintainer=', '.join(['cxrodgers','aiporre','vncntprvst']),
   packages=['WhiskiWrap','wwutils'],
   # scripts=['bin/script1','bin/script2'],
   url='http://pypi.python.org/pypi/WhiskiWrap/',
   license='LICENSE.txt',
   description='Whisk package wrapper created by cxrodgers',
   long_description=open('README.md').read(),
   long_description_content_type='text/markdown',
   install_requires=[
      'tables>=3.5.1',
      'pandas',
      'MediaInfo',
      'future',
      'tifffile',
      'imagecodecs',
      'statsmodels',
      'ffmpeg-python>=0.2.0',
      'whisk-janelia[ffmpeg]'
      ],
   include_package_data=True,
)


#!/usr/bin/env python

import os
import stat
import distutils.sysconfig

def update_permissions():
    # Locate the 'bin' directory in the site-packages
    site_packages_dir = distutils.sysconfig.get_python_lib()
    bin_dir = os.path.join(site_packages_dir, 'whisk', 'bin')

    if not os.path.exists(bin_dir):
        print("Error: 'bin' directory not found. Please ensure the package is installed correctly.")
        return

    #  Test if trace can be executed
    trace_file = os.path.join(bin_dir, 'trace')
    if not os.access(trace_file, os.X_OK):
        print("Updating permissions of executable files in the 'whisk/bin' directory")
        for filename in os.listdir(bin_dir):
            file_path = os.path.join(bin_dir, filename)

            if os.path.isfile(file_path):
                # If on Windows skip files that are not executable
                if os.name == 'nt' and not filename.endswith('.exe'):
                    continue
                # If on Linux, skip files that have an extension
                elif os.name == 'posix' and filename.find('.') != -1:
                    continue

                # For current user: read, write, execute; for group and others: read only
                os.chmod(file_path, stat.S_IRWXU | stat.S_IRGRP | stat.S_IROTH)
                # For all users: read, write, execute
                # os.chmod(file_path, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)
                print(f"Updated permissions for {filename}")

        print("Permissions update complete.")

if __name__ == "__main__":
    update_permissions()



""" wfile_io.py

ctypes interface to libwhisk.so
The functionality described here corresponds with that in trace.h.

File I/O
  Load_Whiskers
  Save_Whiskers

Author: Nathan Clack <clackn@janelia.hhmi.org>
Date  : 25 January 2009

Copyright 2010 Howard Hughes Medical Institute.
All rights reserved.
Use is subject to Janelia Farm Research Campus Software Copyright 1.1
license terms (http://license.janelia.org/license/jfrc_copyright_1_1.html).

Copied to WhiskiWrap repository by Vincent Prevosto, 05/2023
"""
import sys,os
from ctypes import *
from ctypes.util import find_library
from numpy import zeros, float32, uint8, array, hypot, arctan2, pi, concatenate, float64, ndarray, int32
from numpy import where, cos, sin, sum
from warnings import warn

import whisk
# import pdb

def get_all_subdirectories(path):
    return [d for d, _, _ in os.walk(path)]
    # return [d for d in os.listdir(path) if os.path.isdir(os.path.join(path, d))]
    
def ensure_ffmpeg_dlls_are_present(whisk_bin_dir):
    ffmpeg_dll_dir = os.path.join(whisk_bin_dir, 'ffmpeg_win64_lgpl_shared')
    
    if not os.path.exists(ffmpeg_dll_dir):
        import requests, zipfile
        print("First-time setup: downloading necessary ffmpeg DLLs. This might take a few minutes...")
        try:
            from whisk import whisk_utils
            whisk_utils.download_and_extract_ffmpeg_dlls()
            
        except requests.exceptions.RequestException as e:
            # Handle exceptions caused by requests library during the download
            print(f"Error downloading ffmpeg DLLs: {str(e)}")
            print("Please check your internet connection and try again.")
        
        except zipfile.BadZipFile:
            # Handle issues with extraction due to a corrupted download
            print("Error extracting ffmpeg DLLs. The downloaded file might be corrupted.")
            print("Please try running the script again.")
        
        except PermissionError:
            # Handle potential write permission errors gracefully
            print("Permission error while trying to save ffmpeg DLLs. Please ensure you have write permissions to the target directory and try again.")
        
        except Exception as e:
            # Catch all other unexpected exceptions
            print(f"An unexpected error occurred: {str(e)}")
            print("Please report this issue to the software maintainers.")

def load_ffmpeg_dlls(whisk_bin_dir):
    """
    Load ffmpeg DLLs required for cWhisk.
    
    Args:
    - whisk_bin_dir: the imported whisk module's bin directory
    """
    ensure_ffmpeg_dlls_are_present(whisk_bin_dir)

    ffmpeg_dll_dir = os.path.join(whisk_bin_dir, 'ffmpeg_win64_lgpl_shared')

    ffmpeg_dll_names = [
        "avcodec-60.dll",
        "avdevice-60.dll",
        "avformat-60.dll",
        "avutil-58.dll",
        "swscale-7.dll",
    ]

    ffmpeg_dlls = [os.path.join(ffmpeg_dll_dir, dll_name) for dll_name in ffmpeg_dll_names]

    # Load each ffmpeg DLL
    for dll in ffmpeg_dlls:
        CDLL(dll)
        # print(f"Loaded {dll} successfully!")

# Find the base directory of the whisk package
whisk_base_dir = os.path.dirname(whisk.__file__)
whisk_bin_dir = os.path.join(whisk_base_dir, 'bin')
# Set WHISKPATH environment variable to whisk/bin
os.environ['WHISKPATH'] = whisk_bin_dir
# Get all subdirectories of whisk/bin
all_directories = get_all_subdirectories(whisk_bin_dir)

if sys.platform == 'win32':
    lib = os.path.join(whisk_bin_dir, 'whisk.dll')
else:
    lib = os.path.join(whisk_bin_dir, 'libwhisk.so')
    
# Append both whisk base directory and bin directory to PATH
os.environ['PATH'] += os.pathsep + os.pathsep.join(['.', '..', whisk_base_dir, whisk_bin_dir]) 
# Append all bin sub_directories to PATH
os.environ['PATH'] = os.environ['PATH'] + ';' + ';'.join(all_directories)

name = lib
if not name:
  name = find_library('whisk')

try:
    cWhisk = CDLL(name)
except OSError:
    # If loading cWhisk failed, attempt to load ffmpeg DLLs and then try again
    load_ffmpeg_dlls(whisk_bin_dir)
    cWhisk = CDLL(name)

_param_file = "default.parameters"
if cWhisk.Load_Params_File(_param_file)==1: #returns 0 on success, 1 on failure
  cWhisk.Print_Params_File(_param_file)
  if cWhisk.Load_Params_File(_param_file)==1: #returns 0 on success, 1 on failure
    raise Exception("Could not load tracing parameters from file: %s"%_param_file)

#
# DATA STRUCTURE TRANSLATIONS
#

class cContour(Structure):                       #typedef struct
  _fields_ = [( "length"   , c_int            ), #  { int  length;
              ( "boundary" , c_int            ), #    int  boundary;
              ( "width"    , c_int            ), #    int  width;
              ( "iscon4"   , c_int            ), #    int  iscon4;
              ( "tour"     , POINTER( c_int ) )] #    int *tour;
                                                 #  } Contour;

  def asarray(self):
    a = zeros( (self.length,2) )
    for i in range(self.length):
      a[i,0] = self.tour[i]%self.width
      a[i,1] = self.tour[i]/self.width
    return a

  def plot(self,*args,**kwargs):
    from pylab import plot
    a = self.asarray()
    plot(a[:,0],a[:,1],*args,**kwargs)

  def draw(self, surface, color, scale, drawfunc):
    a = self.asarray()
    drawfunc( surface, color, 1, a*scale )

class cObject_Map(Structure):
  _fields_ = [( "num_objects",  c_int ),
              ( "objects",      POINTER(POINTER( cContour ))) ]

  def plot(self,*args,**kwargs):
    for i in range( self.num_objects ):
      self.objects[i].contents.plot(*args,**kwargs)

  def plot_with_seeds( self, image, *args, **kwargs ):
    from pylab import imshow, cm, axis, subplots_adjust, show
    imshow( image, cmap = cm.gray, hold = 0, interpolation = 'nearest' )
    self.plot( *args, **kwargs )
    for i in range(self.num_objects):
      sds = find_seeds( self.objects[i], image )
      if sds:
        sds.plot( linewidths = (1,),
                  facecolors = ('w',),
                  edgecolors = ('k',) )
    axis('image')
    axis('off')
    subplots_adjust(0,0,1,1,0,0)
    show()
    return gcf()

  def draw(self, surface, color, scale, drawfunc ):
    for i in range( self.num_objects ):
      self.objects[i].contents.draw(surface,color,scale,drawfunc)

class cWhisker_Seg_Old(Structure):                 #typedef struct
  _fields_ = [( "id"       , c_int   ),            #  { int    id;
              ( "width"    , c_double),            #    double width;
              ( "beg"      , c_int   ),            #    int    beg;
              ( "end"      , c_int   ),            #    int    end;
              ( "time"     , c_int   ),            #    int    time;
              ( "track"    , POINTER( c_float ) ), #    float *track;
              ( "scores"   , POINTER( c_float ) )] #    float *scores;
                                                  #  } Whisker_Seg_Old;

class cWhisker_Seg(Structure):                      #typedef struct
  _fields_ = [( "id"       , c_int   ),            #{ int id;
              ( "time"     , c_int   ),            #  int time;
              ( "len"      , c_int   ),            #  int len;
              ( "x"        , POINTER( c_float ) ), #  float *x;
              ( "y"        , POINTER( c_float ) ), #  float *y;
              ( "thick"    , POINTER( c_float ) ), #  float *thick;
              ( "scores"   , POINTER( c_float ) )] #  float *scores;
                                                   #} Whisker_Seg;
  @staticmethod
  def CastFromWhiskerSeg( w ):
    return cWhisker_Seg( w.id,
                         w.time,
                         len(w.x),
                         w.x.ctypes.data_as( POINTER( c_float ) ),
                         w.y.ctypes.data_as( POINTER( c_float ) ),
                         w.thick.ctypes.data_as( POINTER( c_float ) ),
                         w.scores.ctypes.data_as( POINTER( c_float ) ) )
  @staticmethod
  def CastDictToArray( wvd ):
    """
    This function creates a ctypes array of cWhisker_Seg from an input
    whiskers dictionary.  The resulting array can be passed to functions
    in the C library.  Don't use Free_Whisker_Seg_Vec on this array; it
    doesn't "own" its data.  Instead, it passes most of the data as pointers
    to numpy arrays.

    Arguments:
    ---------
    <wvd> is a dictionary.  The keys are "frame ids" (integers indicating
          which time point in the movie).  The value's are another set of
          dictionaries.  The keys of these dictionaries are "whisker ids,"
          and the values are of <class `Whisker_Seg`>.

    Example:
    -------
    >>> import trace
    >>> wvd = trace.Load_Whiskers(r"proc\whisker_data_0140.whiskers")
    >>> wv = trace.cWhisker_Seg.CastDictToArray(wvd)
    >>> len(wv)
    40408
    >>> wv[0]
    <trace.cWhisker_Seg object at 0x03DACC60>
    """
    #first count the number of segments
    nseg = sum( map(len, list(wvd.values())) )

    #create the constructor
    type_wv = cWhisker_Seg * nseg;

    #alloc and fill the array
    def itersegs(wvd):
      for v in list(wvd.values()):
        for w in list(v.values()):
          yield w
    wv = type_wv( *list(map(cWhisker_Seg.CastFromWhiskerSeg,list(itersegs(wvd)))) )

    return wv

class cWhiskerIndex(Structure):
  _fields_ = [( 'index' ,   POINTER(POINTER( cWhisker_Seg )) ),
              ( 'sz'    ,   POINTER( c_int  ) ),
              ( 'n'     ,   c_int           ) ]

class cSeed(Structure):
  _fields_ = [( 'xpnt',   c_int ),
              ( 'ypnt',   c_int ),
              ( 'xdir',   c_int ),
              ( 'ydir',   c_int )]

  def asarray(self):
    return array([self.xpnt,self.ypnt,self.xdir,self.ydir])

  def __repr__(self):
    return "<%s.%s (%d,%d) slope: %5.3g>"%(
                str(self.__module__),
                "cSeed",
                self.xpnt,
                self.ypnt,
                arctan2(self.ydir, self.xdir)*180/pi)

class cSeedVector(Structure):                  # typedef struct
  _fields_ = [('nseeds', c_int            ),   #   { int   nseeds;
              ('seeds' , POINTER( cSeed ) )]   #     Seed *seeds;
                                               #   } Seed_Vector;
  def asarray(self):
    a = zeros(( self.nseeds, 4))
    for i in range( self.nseeds ):
      a[i] = self.seeds[i].asarray()
    return a

  def plot(self, *args, **kwargs):
    from pylab import quiver
    a = self.asarray()
    if not 'pivot' in kwargs:
      kwargs['pivot'] = 'middle'
    if not 'scale' in kwargs:
      kwargs['scale'] = 10
    if a.any():
      norm = hypot( a[:,2], a[:,3] )
      a[:,2] /= norm
      a[:,3] /= norm
      quiver( a[:,0], a[:,1], a[:,2], -a[:,3], **kwargs )

class cImage(Structure):
  _fields_ = [( "kind",     c_int   ),
              ( "width",    c_int   ),
              ( "height",   c_int   ),
              ( "text",     POINTER(c_char_p) ),
              ( "array",    POINTER(c_uint8))]
  @staticmethod
  def fromarray(im):
    return cImage(  c_int(  im.dtype.itemsize ),
                    c_int(  im.shape[1] ),
                    c_int(  im.shape[0] ),
                    pointer( c_char_p("") ),
                    im.ctypes.data_as( POINTER( c_uint8 ) ) )

class Whisker_Seg(object):
  def __init__(self, source=None):
    """
    If source is:
      None        : initialize an empty segment
      cWhisker_Seg: copy contents of cWhisker_Seg into a Whisker_Seg
      tuple       : must be (Whisker_Seg, start_index, stop_index)
                    Create a Whisker_Seg referencng the indicated subsegment

    """
    if source is None:
      self.id   = None
      self.time = None
      self.x    = array([])
      self.y    = array([])
      self.scores = array([])
      self.thick  = array([])
    elif isinstance( source, cWhisker_Seg ):
      self.id = source.id
      self.time = source.time

      self.x      = zeros( source.len, dtype=float32 )
      self.y      = zeros( source.len, dtype=float32)
      self.scores = zeros( source.len, dtype=float32 )
      self.thick  = zeros( source.len, dtype=float32 )

      for i in range( source.len ):
        self.x[i]      = source.x[i]
        self.y[i]      = source.y[i]
        self.thick[i]  = source.thick[i]
        self.scores[i] = source.scores[i]
    elif isinstance( source, tuple ):
      arg,a,b = source
      self.id   = arg.id
      self.time = arg.time
      self.x      = arg.x[a:b]
      self.y      = arg.y[a:b]
      self.thick  = arg.thick[a:b]
      self.scores = arg.scores[a:b]

  def __len__(self):
    return len(self.x)

  def __getitem__(self, idx ):
    return [self.x[idx],self.y[idx],self.thick[idx],self.scores[idx]]

  def asarray( self ):
    return concatenate((self.x,self.y,self.thick,self.scores)).reshape(4,len(self.x)).T

  def split( self, idx ):
    a,b = Whisker_Seg(), Whisker_Seg()
    a.id   = b.id   = self.id
    a.time = b.time = self.time
    a.x      = self.x     [:idx]
    a.y      = self.y     [:idx]
    a.thick  = self.thick [:idx]
    a.scores = self.scores[:idx]
    b.x      = self.x     [idx:]
    b.y      = self.y     [idx:]
    b.thick  = self.thick [idx:]
    b.scores = self.scores[idx:]
    if len(a) == 0:
      a = None
    if len(b) == 0:
      b = None
    return a,b

  def join_right( self, w ):
    self.x      = concatenate( (self.x     , w.x     ) )
    self.y      = concatenate( (self.y     , w.y     ) )
    self.thick  = concatenate( (self.thick , w.thick ) )
    self.scores = concatenate( (self.scores, w.scores) )
    return self

  @staticmethod
  def join( left, right ):
    """ Like `join_right` but the joined result is a new whisker. """
    w = left.copy()
    w.join_right( right )
    return w

  def copy(self):
    w = Whisker_Seg()
    w.id = self.id
    w.time = self.time

    w.x      = self.x.copy()
    w.y      = self.y.copy()
    w.scores = self.scores.copy()
    w.thick  = self.thick.copy()
    return w

#
# FILE I/O
#

cWhisk.Load_Whiskers.restype = POINTER( cWhisker_Seg )
cWhisk.Load_Whiskers.argtypes = [
  POINTER( c_char ),
  POINTER( c_char ),
  POINTER( c_int)]

cWhisk.Save_Whiskers.restype = c_int
cWhisk.Save_Whiskers.argtypes = [
  POINTER( c_char ),
  POINTER( c_char ),
  POINTER( cWhisker_Seg ),
  c_int]

def Load_Whiskers( filename ):
  """ Reads whisker segments from a file.

  Returns a dict of dict with Whisker_Seg elements.  The organization looks
  like:

  >>> from trace import Load_Whiskers
  >>> frameid = 0;
  >>> segmentid = 1;
  >>> wv = Load_Whiskers('test.whiskers')
  >>> w = wv[frameid][segmentid]
  >>> w.id == segmentid
  True
  >>> w.time == frameid
  True

  The file is read using a ctypes call to a libwhisk(trace.h) function.  The
  data is then copied into numpy containers and freed.

  It's not clear to me how this effects the part of the heap hidden from
  python's memmory management.  This is done for convenience, since later no
  one needs to remember to free anything and one gets to use numpy.array's for
  plotting, etc...

  For the kinds of things I imaging doing in python this function will only get
  called once per application instance.
  """
  if not os.path.exists(filename):
    raise IOError("File not found.")
  nwhiskers = c_int(0)
  filenameb = bytes(filename, 'utf-8')
  wv = cWhisk.Load_Whiskers( filenameb, None, byref(nwhiskers) );
  # organize into dictionary for ui.py {frameid}{segid}
  whiskers = {}
  for idx in range( nwhiskers.value ):
    w = wv[idx]
    whiskers[ w.time ] = {}
  for idx in range( nwhiskers.value ):
    w = Whisker_Seg(wv[idx])
    whiskers[ w.time ][ w.id ] = w;
  cWhisk.Free_Whisker_Seg_Vec( wv, nwhiskers )
  return whiskers

def Save_Whiskers( filename, whiskers ):
  #count the whiskers
  n = 0
  for v in list(whiskers.values()):
    n += len(v)
  #alloc the c whisker array
  wv = (cWhisker_Seg * n)()
  #copy into c whisker array
  i = 0
  for fid,v in list(whiskers.items()):
    for wid,t in list(v.items()):
      if not t:
        continue;
      wv[i].id   = wid
      wv[i].time = fid
      wv[i].len  = len(t.x)
      wv[i].x       = t.x.ctypes.data_as( POINTER( c_float ) )
      wv[i].y       = t.y.ctypes.data_as( POINTER( c_float ) )
      wv[i].thick   = t.thick.ctypes.data_as( POINTER( c_float ) )
      wv[i].scores  = t.scores.ctypes.data_as( POINTER( c_float ) )
      i += 1
  #prep for save and save
  #print "Saving %d"%i
  #pdb.set_trace()
  if not cWhisk.Save_Whiskers( filename, None, wv, i ):
    warn("Save Whiskers may have failed.")




"""Main functions for running input videos through trace.

The overall algorithm is contained in `interleaved_reading_and_tracing`.
* The input can be a video file or a directory of PF files.
* Chunks of ~200 frames are read using ffmpeg, and then written to disk
  as uncompressed tiff stacks.
* Trace is called in parallel on each tiff stack
* Additional chunks are read as trace completes.
* At the end, all of the HDF5 files are stitched together.

The previous function `pipeline_trace` is now deprecated.
"""
import traceback
import sys
try:
    import tifffile
except ImportError:
    pass
import os
import re
import numpy as np
import pandas as pd
import subprocess
import multiprocessing
import tables
# from . import wfile_io
# from .mfile_io import MeasurementsTable
# for debugging
from WhiskiWrap import wfile_io
from WhiskiWrap.mfile_io import MeasurementsTable

# try:
#     from whisk import trace
#     from whisk.traj import MeasurementsTable
# except ImportError as e:
#     traceback.print_tb(e.__traceback__)
#     print("cannot import whisk")
if sys.platform == 'win32':
    whisk_path = ''
else:
    whisk_path = os.environ['WHISKPATH']
    if not whisk_path.endswith('/'):
        whisk_path += '/'
    print('WHISKPATH detected: ', whisk_path)
import pandas
import WhiskiWrap
from WhiskiWrap import video_utils
import wwutils
import scipy.io
import ctypes
import glob
import time
import shutil
import itertools

# Find the repo directory and the default param files
# The banks don't differe with sensitive or default
DIRECTORY = os.path.split(__file__)[0]
PARAMETERS_FILE = os.path.join(DIRECTORY, 'default.parameters')
SENSITIVE_PARAMETERS_FILE = os.path.join(DIRECTORY, 'sensitive.parameters')
HALFSPACE_DB_FILE = os.path.join(DIRECTORY, 'halfspace.detectorbank')
LINE_DB_FILE = os.path.join(DIRECTORY, 'line.detectorbank')

# libpfDoubleRate library, needed for PFReader
LIB_DOUBLERATE = os.path.join(DIRECTORY, 'libpfDoubleRate.so')

def copy_parameters_files(target_directory, sensitive=False):
    """Copies in parameters and banks"""
    if sensitive:
        shutil.copyfile(SENSITIVE_PARAMETERS_FILE, os.path.join(target_directory,
            'default.parameters'))
    else:
        shutil.copyfile(PARAMETERS_FILE, os.path.join(target_directory,
            'default.parameters'))

    # Banks are the same regardless
    shutil.copyfile(HALFSPACE_DB_FILE, os.path.join(target_directory,
        'halfspace.detectorbank'))
    shutil.copyfile(LINE_DB_FILE, os.path.join(target_directory,
        'line.detectorbank'))

class WhiskerSeg(tables.IsDescription):
    fid = tables.UInt32Col()
    wid = tables.UInt16Col()
    tip_x = tables.Float32Col()
    tip_y = tables.Float32Col()
    follicle_x = tables.Float32Col()
    follicle_y = tables.Float32Col()
    pixel_length = tables.UInt16Col()
    chunk_start = tables.UInt32Col()

class WhiskerSeg_measure(tables.IsDescription):
    fid = tables.UInt32Col()
    wid = tables.UInt16Col()
    tip_x = tables.Float32Col()
    tip_y = tables.Float32Col()
    follicle_x = tables.Float32Col()
    follicle_y = tables.Float32Col()
    pixel_length = tables.UInt16Col()
    length = tables.Float32Col()
    score = tables.Float32Col()
    angle = tables.Float32Col()
    curvature = tables.Float32Col()
    chunk_start = tables.UInt32Col()
    # Adding previous missing fields label, face_x, face_y:
    label = tables.UInt16Col()
    # face_x and face_y are signed integers as the face location assigned by measure may be outside the frame
    face_x = tables.Int32Col()
    face_y = tables.Int32Col()

def write_chunk(chunk, chunkname, directory='.'):
    tifffile.imsave(os.path.join(directory, chunkname), chunk, compress=0)

def trace_chunk(video_filename, delete_when_done=False):
    """Run trace on an input file

    First we create a whiskers filename from `video_filename`, which is
    the same file with '.whiskers' replacing the extension. Then we run
    trace using subprocess.

    Care is taken to move into the working directory during trace, and then
    back to the original directory.

    Returns:
        stdout, stderr
    """
    print(("Starting", video_filename))
    orig_dir = os.getcwd()
    run_dir, raw_video_filename = os.path.split(os.path.abspath(video_filename))
    whiskers_file = WhiskiWrap.utils.FileNamer.from_video(video_filename).whiskers
    command = [whisk_path + 'trace', raw_video_filename, whiskers_file]

    os.chdir(run_dir)
    try:
        pipe = subprocess.Popen(command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            )
        stdout, stderr = pipe.communicate()
    except:
        raise
    finally:
        os.chdir(orig_dir)
    print(("Done", video_filename))

    if not os.path.exists(whiskers_file):
        print(raw_video_filename)
        raise IOError("tracing seems to have failed. STDOUT: " + stdout.decode('ascii') + '. STDERR: ' + stderr.decode('ascii')  )

    if delete_when_done:
        os.remove(video_filename)

    return {'video_filename': video_filename, 'stdout': stdout, 'stderr': stderr}

def measure_chunk(whiskers_filename, face, delete_when_done=False):
    """Run measure on an input file

    First we create a measurement filename from `whiskers_filename`, which is
    the same file with '.measurements' replacing the extension. Then we run
    trace using subprocess.

    Care is taken to move into the working directory during trace, and then
    back to the original directory.

    Returns:
        stdout, stderr
    """
    print(("Starting", whiskers_filename))
    orig_dir = os.getcwd()
    run_dir, raw_whiskers_filename = os.path.split(os.path.abspath(whiskers_filename))
    measurements_file = WhiskiWrap.utils.FileNamer.from_whiskers(whiskers_filename).measurements
    command = ['measure', '--face', face, raw_whiskers_filename, measurements_file]

    os.chdir(run_dir)
    try:
        pipe = subprocess.Popen(command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            )
        stdout, stderr = pipe.communicate()
    except:
        raise
    finally:
        os.chdir(orig_dir)
    print(("Done", whiskers_filename))

    if not os.path.exists(measurements_file):
        print(raw_whiskers_filename)
        raise IOError("measurement seems to have failed")

    if delete_when_done:
        os.remove(whiskers_filename)

    return {'whiskers_filename': whiskers_filename, 'stdout': stdout, 'stderr': stderr}

def trace_and_measure_chunk(video_filename, delete_when_done=False, face='right', classify=None):
    """Run trace on an input file

    First we create a whiskers filename from `video_filename`, which is
    the same file with '.whiskers' replacing the extension. Then we run
    trace using subprocess.

    Care is taken to move into the working directory during trace, and then
    back to the original directory.

    Returns:
        stdout, stderr
    """
    print("Starting", video_filename)

    orig_dir = os.getcwd()
    run_dir, raw_video_filename = os.path.split(os.path.abspath(video_filename))

    # Run trace:
    whiskers_file = WhiskiWrap.utils.FileNamer.from_video(video_filename).whiskers
    trace_command = [whisk_path + 'trace', raw_video_filename, whiskers_file]

    os.chdir(run_dir)
    try:
        trace_pipe = subprocess.Popen(trace_command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            )
        stdout, stderr = trace_pipe.communicate()
    except Exception as e:
        print(e)
        raise
    finally:
        os.chdir(orig_dir)
    print("Done", video_filename)

    if not os.path.exists(whiskers_file):
        print(raw_video_filename)
        raise IOError("tracing seems to have failed")

    # Run measure:
    measurements_file = WhiskiWrap.utils.FileNamer.from_video(video_filename).measurements
    measure_command = ['measure', '--face', face, whiskers_file, measurements_file]

    os.chdir(run_dir)
    try:
        measure_pipe = subprocess.Popen(measure_command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            )
        stdout, stderr = measure_pipe.communicate()
    except Exception as e:
        print(e)
        raise
    finally:
        os.chdir(orig_dir)
    print("Done", whiskers_file)

    if not os.path.exists(measurements_file):
        print(whiskers_file)
        raise IOError("measuring seems to have failed")
    
    # Run classify / re-classify:
    # typical classify arguments are {'px2mm': '0.04', 'n_whiskers': '3'}
    if classify is not None:
        measurements_file = WhiskiWrap.utils.FileNamer.from_video(video_filename).measurements
        classify_command = ['classify', measurements_file, measurements_file, 
                            face, '--px2mm', classify['px2mm'], '-n', classify['n_whiskers']]
        reclassify_command = ['reclassify', measurements_file, measurements_file, '-n', classify['n_whiskers']]

        os.chdir(run_dir)
        try:
            classify_pipe = subprocess.Popen(classify_command,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                )
            stdout, stderr = classify_pipe.communicate()
            reclassify_pipe = subprocess.Popen(reclassify_command,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                )
            stdout, stderr = reclassify_pipe.communicate()
        except Exception as e:
            print(e)
            raise
        finally:
            os.chdir(orig_dir)
        print("Done", measurements_file)

        if not os.path.exists(measurements_file):
            print(measurements_file)
            raise IOError("classify seems to have failed")

    # Clean up:
    if delete_when_done:
        os.remove(video_filename)

    return {'video_filename': video_filename,'stdout': stdout, 'stderr': stderr}

def sham_trace_chunk(video_filename):
    print(("sham tracing", video_filename))
    time.sleep(2)
    return video_filename

def setup_hdf5(h5_filename, expectedrows, measure=False):

    # Open file
    h5file = tables.open_file(h5_filename, mode="w")

    if not measure:
        WhiskerDescription = WhiskerSeg
    elif measure:
        WhiskerDescription = WhiskerSeg_measure

    # A group for the normal data
    table = h5file.create_table(h5file.root, "summary", WhiskerDescription,
        "Summary data about each whisker segment",
        expectedrows=expectedrows)

    # Put the contour here
    xpixels_vlarray = h5file.create_vlarray(
        h5file.root, 'pixels_x',
        tables.Float32Atom(shape=()),
        title='Every pixel of each whisker (x-coordinate)',
        expectedrows=expectedrows)
    ypixels_vlarray = h5file.create_vlarray(
        h5file.root, 'pixels_y',
        tables.Float32Atom(shape=()),
        title='Every pixel of each whisker (y-coordinate)',
        expectedrows=expectedrows)

    h5file.close()

def index_measurements(whiskers,measurements):
    print("classify was run on the measurements file")

    # We get all the whisker ids from the whiskers dictionary. Since ids are not unique (they repeat for each frame),
    # we also get the frame number, to get a unique set of identifiers. 
    # whisker_id_set = [] 
    # for frame, frame_whiskers in list(whiskers.items()):
    #     for whisker_id, wseg in list(frame_whiskers.items()):
    #         whisker_id_set.append([frame, wseg.id])

    # # Convert to set for faster lookup
    # whisker_id_set = set(whisker_id_set)
    whisker_id_set = set() 
    for frame, frame_whiskers in whiskers.items():
        for whisker_id, wseg in frame_whiskers.items():
            whisker_id_set.add((frame, wseg.id))

    # Then, get all the same data from the measurements dictionary (converting to int)
    measurement_ids = []
    for measurement in measurements:
        measurement_ids.append((int(measurement[1]), int(measurement[2])))

    # Convert to dictionary for faster lookup
    measurement_ids_dict = {id: idx for idx, id in enumerate(measurement_ids)}

    # Now, match the whisker ids from the whiskers dictionary to the whisker ids from the measurements dictionary
    # This is done by matching the frame number and the whisker id
    # The result is a list of indices that can be used to index the measurements array
    measurements_reidx = []
    for whisker_id in whisker_id_set:
        idx = measurement_ids_dict.get(whisker_id, -1)
        measurements_reidx.append(idx)
    # for whisker_id in whisker_id_set:
    #     for idx, measurement_id in enumerate(measurement_ids):
    #         if measurement_id == whisker_id:
    #             measurements_reidx.append(idx)
    #             break
    #     # if the whisker was not found in the measurements array, then it was removed by classify
    #     # so we add a -1 to the measurements_reidx list
    #     if measurement_id != whisker_id:
    #         measurements_reidx.append(-1)

    # Find the indices of the measurements that are -1
    removed_idx = [i for i, x in enumerate(measurements_reidx) if x == -1]
    print("Number of whiskers removed by classify: ", len(removed_idx))

    # Now, we can index the measurements array using the indices in measurements_reidx
    measurements = measurements[measurements_reidx]
    
    return measurements
              
def append_whiskers_to_hdf5(whisk_filename, h5_filename, chunk_start, measurements_filename=None, summary_only=False):
    """Load data from whisk_file and put it into an hdf5 file

    The HDF5 file will have two basic components:
        /summary : A table with the following columns:
            fid, wid, follicle_x, follicle_y, tip_x, tip_y, pixel_length
            These are all directly taken from the whisk file
        /pixels_x : A vlarray of the same length as summary but with the
            entire array of x-coordinates of each segment.
        /pixels_y : Same but for y-coordinates
    """
    ## Load it, so we know what expected rows is
    # This loads all whisker info into C data types
    # wv is like an array of trace.LP_cWhisker_Seg
    # Each entry is a trace.cWhisker_Seg and can be converted to
    # a python object via: wseg = trace.Whisker_Seg(wv[idx])
    # The python object responds to .time and .id (integers) and .x and .y (numpy
    # float arrays).
    #wv, nwhisk = trace.Debug_Load_Whiskers(whisk_filename)
    print(whisk_filename)

    whiskers = wfile_io.Load_Whiskers(whisk_filename)
    # whiskers is a list of dictionaries, each dictionary is a frame, each frame has a dictionary of whiskers
    # len(whiskers) is = to chunk_size (which is a number of frames)

    nwhisk = np.sum(list(map(len, list(whiskers.values()))))

    if measurements_filename is not None:
        print(measurements_filename)
        M = MeasurementsTable(str(measurements_filename))
        measurements = M.asarray()
        # len(measurements) is = to number of whiskers in those frames = chunk_size * n whiskers per frame

        measurements_idx = 0

        # If `classify` was run on the measurements file, then the index of the individual whiskers will have changed
        # Basically, the order of measurements is scrambled with respect to wseg.
        # Also, some whiskers may have been removed.
        # So we need to match wseg.id to measurements[2]

        # First check whether classify was run on the measurements file, by comparing whisker ids in the first frame
        # of the whiskers dictionary to the whisker ids in the measurements array

        wid_from_trace = np.array(list(whiskers[0].keys())).astype(int)
        initial_frame_measurements = measurements[:len(wid_from_trace)]
        wid_from_measure = initial_frame_measurements[:, 2].astype(int)

        if not np.array_equal(wid_from_trace, wid_from_measure):
            measurements=index_measurements(whiskers,measurements)
            
    # Open file
    h5file = tables.open_file(h5_filename, mode="a")

    # See setup_hdf5 for creation of the table
    # 05/11/23: changed WhiskerSeg_measure to add face_x / face_y / label fields

    ## Iterate over rows and store
    table = h5file.get_node('/summary')
    h5seg = table.row
    if not summary_only:
        xpixels_vlarray = h5file.get_node('/pixels_x')
        ypixels_vlarray = h5file.get_node('/pixels_y')
    for frame, frame_whiskers in list(whiskers.items()):
        for whisker_id, wseg in list(frame_whiskers.items()):
            # Write to the table
            h5seg['chunk_start'] = chunk_start
            h5seg['fid'] = wseg.time + chunk_start
            h5seg['wid'] = wseg.id

            if measurements_filename is not None:
                h5seg['length'] = measurements[measurements_idx][3]
                h5seg['score'] = measurements[measurements_idx][4]
                h5seg['angle'] = measurements[measurements_idx][5]
                h5seg['curvature'] = measurements[measurements_idx][6]
                h5seg['pixel_length'] = len(wseg.x)
                h5seg['follicle_x'] = measurements[measurements_idx][7]
                h5seg['follicle_y'] = measurements[measurements_idx][8]
                h5seg['tip_x'] = measurements[measurements_idx][9]
                h5seg['tip_y'] = measurements[measurements_idx][10]
                h5seg['label'] = 0
                h5seg['face_x'] = M._measurements.contents.face_x
                h5seg['face_y'] = M._measurements.contents.face_y

                measurements_idx += 1

            else:
                h5seg['follicle_x'] = wseg.x[-1]
                h5seg['follicle_y'] = wseg.y[-1]
                h5seg['tip_x'] = wseg.x[0]
                h5seg['tip_y'] = wseg.y[0]
   
            assert len(wseg.x) == len(wseg.y)
            h5seg.append()

            if not summary_only:
            # Write whisker contour x and y pixel values
                xpixels_vlarray.append(wseg.x)
                ypixels_vlarray.append(wseg.y)

    table.flush()
    h5file.close()

def initialize_whisker_measurement_table():
    """Initialize tabular data to enter into WhiskerMeasurementTable, using Pandas DataFrames"""
    # Define column types
    column_types = {
        'frame_id': 'uint32',
        'whisker_id': 'uint16',
        'label': 'uint16',
        'face_x': 'int32',
        'face_y': 'int32',
        'length': 'float32',
        'pixel_length': 'uint16',
        'score': 'float32',
        'angle': 'float32',
        'curvature': 'float32',
        'follicle_x': 'float32',
        'follicle_y': 'float32',
        'tip_x': 'float32',
        'tip_y': 'float32',
        'chunk_start': 'uint32'
    }
    
    measurement_data = pd.DataFrame(columns=column_types.keys()).astype(column_types)

    return measurement_data, column_types

def read_whisker_data(filename, output_format='dict'):
    """
    Parameter: filename (str) - path to whisker file.
    Accepts hdf5 and .whiskers files. If the latter, it will look for a .measurements file in the same directory.
    Returns a dictionary with the following keys:
        'whisker_id' - list of whisker ids
        'frame_id' - list of frame ids
        'follicle_x' - list of follicle x coordinates
        'follicle_y' - list of follicle y coordinates
        'tip_x' - list of tip x coordinates
        'tip_y' - list of tip y coordinates
        'pixel_length' - list of pixel lengths
        'length' - list of lengths
        'score' - list of scores
        'angle' - list of angles
        'curvature' - list of curvatures
        'face_x' - list of face x coordinates
        'face_y' - list of face y coordinates
        'label' - list of labels
        'chunk_start' - list of chunk start indices
    """    
    if filename is None:
        print("No measurements file provided")
        return None
    
    if filename.endswith('.whiskers'):
        # Read whisker file
        whiskers = wfile_io.Load_Whiskers(filename)
        
        # check if measurements file exists
        measurements_filename = filename.replace('.whiskers', '.measurements')
        
        if os.path.isfile(measurements_filename):
            # print(measurements_filename)
            M = MeasurementsTable(str(measurements_filename))
            measurements = M.asarray()
            measurements_idx = 0
        else:
            # return in error
            print("No measurements file found")
            return None
            
        # First check whether classify was run on the measurements file, by comparing whisker ids in the first frame of the whiskers dictionary to the whisker ids in the measurements array

        wid_from_trace = np.array(list(whiskers[0].keys())).astype(int)
        initial_frame_measurements = measurements[:len(wid_from_trace)]
        wid_from_measure = initial_frame_measurements[:, 2].astype(int)

        if not np.array_equal(wid_from_trace, wid_from_measure):
            measurements=index_measurements(whiskers,measurements)

        meas_rows = []
        chunk_start = 0
        
        ## Iterate over rows and append to table
        for frame, frame_whiskers in list(whiskers.items()):
            for whisker_id, wseg in list(frame_whiskers.items()):
                # Write to the table, using Pandas DataFrame (more efficient for adding multiple rows):
                # Create a new row as a dictionary
                new_meas_row = {
                    'frame_id': wseg.time + chunk_start,
                    'whisker_id': wseg.id,
                    'label': 0,
                    'face_x': M._measurements.contents.face_x,
                    'face_y': M._measurements.contents.face_y,
                    'length': measurements[measurements_idx][3],
                    'pixel_length': len(wseg.x),
                    'score': measurements[measurements_idx][4],
                    'angle': measurements[measurements_idx][5],
                    'curvature': measurements[measurements_idx][6],
                    'follicle_x': measurements[measurements_idx][7],
                    'follicle_y': measurements[measurements_idx][8],
                    'tip_x': measurements[measurements_idx][9],
                    'tip_y': measurements[measurements_idx][10],
                    'chunk_start': chunk_start,
                }
                
                measurements_idx += 1
                assert len(wseg.x) == len(wseg.y)
                
                # Append the new row to the DataFrame
                # meas_table = meas_table.append(new_meas_row, ignore_index=True)
                meas_rows.append(new_meas_row)
                
                # Write whisker contour x and y pixel values
                # %TODO: save whisker contours
                # xpixels_vlarray.append(wseg.x)
                # ypixels_vlarray.append(wseg.y)
                
        # no need to initilize the table, just create a new one
        _, column_types = initialize_whisker_measurement_table()
        meas_table = pd.DataFrame(meas_rows).astype(column_types)
        
    elif filename.endswith('.hdf5'):   
        # Read hdf5 file and returns a pandas DataFrame
        meas_table = read_whiskers_hdf5_summary(filename)
        # print(meas_table.head())

        # Convert non-matching table labels
        meas_table.rename(columns={'fid': 'frame_id', 'wid': 'whisker_id'}, inplace=True)
                
    # Convert to dictionary
    if output_format == 'dict':
        data_dict = {k: meas_table[k].values for k in meas_table.columns}
        return data_dict
    elif output_format == 'df':
        return meas_table

def pipeline_trace(input_vfile, h5_filename,
    epoch_sz_frames=3200, chunk_sz_frames=200,
    frame_start=0, frame_stop=None,
    n_trace_processes=4, expectedrows=1000000, flush_interval=100000,
    measure=False,face='right'):
    """Trace a video file using a chunked strategy.

    This is now deprecated in favor of interleaved_reading_and_tracing.
    The issue with this function is that it has to write out all of the tiffs
    first, before tracing, which is a wasteful use of disk space.

    input_vfile : input video filename
    h5_filename : output HDF5 file
    epoch_sz_frames : Video is first broken into epochs of this length
    chunk_sz_frames : Each epoch is broken into chunks of this length
    frame_start, frame_stop : where to start and stop processing
    n_trace_processes : how many simultaneous processes to use for tracing
    expectedrows, flush_interval : used to set up hdf5 file

    TODO: combine the reading and writing stages using frame_func so that
    we don't have to load the whole epoch in at once. In fact then we don't
    even need epochs at all.
    """
    WhiskiWrap.utils.probe_needed_commands()

    # Figure out where to store temporary data
    input_vfile = os.path.abspath(input_vfile)
    input_dir = os.path.split(input_vfile)[0]

    # Setup the result file
    setup_hdf5(h5_filename, expectedrows, measure=measure)

    # Figure out how many frames and epochs
    duration = wwutils.video.get_video_duration(input_vfile)
    frame_rate = wwutils.video.get_video_params(input_vfile)[2]
    total_frames = int(np.rint(duration * frame_rate))
    if frame_stop is None:
        frame_stop = total_frames
    if frame_stop > total_frames:
        print("too many frames requested, truncating")
        frame_stop = total_frames

    # Iterate over epochs
    for start_epoch in range(frame_start, frame_stop, epoch_sz_frames):
        # How many frames in this epoch
        stop_epoch = np.min([frame_stop, start_epoch + epoch_sz_frames])
        print(("Epoch %d - %d" % (start_epoch, stop_epoch)))

        # Chunks
        chunk_starts = np.arange(start_epoch, stop_epoch, chunk_sz_frames)
        chunk_names = ['chunk%08d.tif' % nframe for nframe in chunk_starts]
        whisk_names = ['chunk%08d.whiskers' % nframe for nframe in chunk_starts]

        # read everything
        # need to be able to crop here
        print("Reading")
        frames = video_utils.process_chunks_of_video(input_vfile,
            frame_start=start_epoch, frame_stop=stop_epoch,
            frames_per_chunk=chunk_sz_frames, # only necessary for chunk_func
            frame_func=None, chunk_func=None,
            verbose=False, finalize='listcomp')

        # Dump frames into tiffs or lossless
        print("Writing")
        for n_whiski_chunk, chunk_name in enumerate(chunk_names):
            print(n_whiski_chunk)
            chunkstart = n_whiski_chunk * chunk_sz_frames
            chunkstop = (n_whiski_chunk + 1) * chunk_sz_frames
            chunk = frames[chunkstart:chunkstop]
            if len(chunk) in [3, 4]:
                print("WARNING: trace will fail on tiff stacks of length 3 or 4")
            write_chunk(chunk, chunk_name, input_dir)

        # Also write lossless and/or lossy monitor video here?
        # would really only be useful if cropping applied

        # trace each
        print("Tracing")
        pool = multiprocessing.Pool(n_trace_processes)
        trace_res = pool.map(trace_chunk,
            [os.path.join(input_dir, chunk_name)
                for chunk_name in chunk_names])
        pool.close()

        # take measurements:
        if measure:
            print("Measuring")
            pool = multiprocessing.Pool(n_trace_processes)
            meas_res = pool.map(measure_chunk_star,
                list(zip([os.path.join(input_dir, whisk_name)
                    for whisk_name in whisk_names],itertools.repeat(face))))
            pool.close()

        # stitch
        print("Stitching")
        for chunk_start, chunk_name in zip(chunk_starts, chunk_names):
            # Append each chunk to the hdf5 file
            fn = WhiskiWrap.utils.FileNamer.from_tiff_stack(
                os.path.join(input_dir, chunk_name))

            if not measure:
                append_whiskers_to_hdf5(
                    whisk_filename=fn.whiskers,
                    h5_filename=h5_filename,
                    chunk_start=chunk_start)
            elif measure:
                append_whiskers_to_hdf5(
                    whisk_filename=fn.whiskers,
                    measurements_filename=fn.measurements,
                    h5_filename=h5_filename,
                    chunk_start=chunk_start)

def write_video_as_chunked_tiffs(input_reader, tiffs_to_trace_directory,
    chunk_size=200, chunk_name_pattern='chunk%08d.tif',
    stop_after_frame=None, monitor_video=None, timestamps_filename=None,
    monitor_video_kwargs=None):
    """Write frames to disk as tiff stacks

    input_reader : object providing .iter_frames() method and perhaps
        also a .timestamps attribute. For instance, PFReader, or some
        FFmpegReader object.
    tiffs_to_trace_directory : where to store the chunked tiffs
    stop_after_frame : to stop early
    monitor_video : if not None, should be a filename to write a movie to
    timestamps_filename : if not None, should be the name to write timestamps
    monitor_video_kwargs : ffmpeg params

    Returns: ChunkedTiffWriter object
    """
    # Tiff writer
    ctw = WhiskiWrap.ChunkedTiffWriter(tiffs_to_trace_directory,
        chunk_size=chunk_size, chunk_name_pattern=chunk_name_pattern)

    # FFmpeg writer is initalized after first frame
    ffw = None

    # Iterate over frames
    for nframe, frame in enumerate(input_reader.iter_frames()):
        # Stop early?
        if stop_after_frame is not None and nframe >= stop_after_frame:
            break

        # Write to chunked tiff
        ctw.write(frame)

        # Optionally write to monitor video
        if monitor_video is not None:
            # Initialize ffw after first frame so we know the size
            if ffw is None:
                ffw = WhiskiWrap.FFmpegWriter(monitor_video,
                    frame_width=frame.shape[1], frame_height=frame.shape[0],
                    **monitor_video_kwargs)
            ffw.write(frame)

    # Finalize writers
    ctw.close()
    if ffw is not None:
        ff_stdout, ff_stderr = ffw.close()

    # Also write timestamps as numpy file
    if hasattr(input_reader, 'timestamps') and timestamps_filename is not None:
        timestamps = np.concatenate(input_reader.timestamps)
        assert len(timestamps) >= ctw.frames_written
        np.save(timestamps_filename, timestamps[:ctw.frames_written])

    return ctw

def trace_chunked_tiffs(input_tiff_directory, h5_filename,
    n_trace_processes=4, expectedrows=1000000,
    ):
    """Trace tiffs that have been written to disk in parallel and stitch.

    input_tiff_directory : directory containing tiffs
    h5_filename : output HDF5 file
    n_trace_processes : how many simultaneous processes to use for tracing
    expectedrows : used to set up hdf5 file
    """
    WhiskiWrap.utils.probe_needed_commands()

    # Setup the result file
    setup_hdf5(h5_filename, expectedrows)

    # The tiffs have been written, figure out which they are
    tif_file_number_strings = wwutils.misc.apply_and_filter_by_regex(
        '^chunk(\d+).tif$', os.listdir(input_tiff_directory), sort=False)
    tif_full_filenames = [
        os.path.join(input_tiff_directory, 'chunk%s.tif' % fns)
        for fns in tif_file_number_strings]
    tif_file_numbers = list(map(int, tif_file_number_strings))
    tif_ordering = np.argsort(tif_file_numbers)
    tif_sorted_filenames = np.array(tif_full_filenames)[
        tif_ordering]
    tif_sorted_file_numbers = np.array(tif_file_numbers)[
        tif_ordering]

    # trace each
    print("Tracing")
    pool = multiprocessing.Pool(n_trace_processes)
    trace_res = pool.map(trace_chunk, tif_sorted_filenames)
    pool.close()

    # stitch
    print("Stitching")
    for chunk_start, chunk_name in zip(tif_sorted_file_numbers, tif_sorted_filenames):
        # Append each chunk to the hdf5 file
        fn = WhiskiWrap.utils.FileNamer.from_tiff_stack(chunk_name)
        append_whiskers_to_hdf5(
            whisk_filename=fn.whiskers,
            h5_filename=h5_filename,
            chunk_start=chunk_start)

def interleaved_read_trace_and_measure(input_reader, tiffs_to_trace_directory,
    sensitive=False,
    chunk_size=200, chunk_name_pattern='chunk%08d.tif',
    stop_after_frame=None, delete_tiffs=True,
    timestamps_filename=None, monitor_video=None,
    monitor_video_kwargs=None, write_monitor_ffmpeg_stderr_to_screen=False,
    h5_filename=None, frame_func=None,
    n_trace_processes=4, expectedrows=1000000,
    verbose=True, skip_stitch=False, face='right'
    ):
    """Read, write, and trace each chunk, one at a time.

    This is an alternative to first calling:
        write_video_as_chunked_tiffs
    And then calling
        trace_chunked_tiffs

    input_reader : Typically a PFReader or FFmpegReader
    tiffs_to_trace_directory : Location to write the tiffs
    sensitive: if False, use default. If True, lower MIN_SIGNAL
    chunk_size : frames per chunk
    chunk_name_pattern : how to name them
    stop_after_frame : break early, for debugging
    delete_tiffs : whether to delete tiffs after done tracing
    timestamps_filename : Where to store the timestamps
        Only vallid for PFReader input_reader
    monitor_video : filename for a monitor video
        If None, no monitor video will be written
    monitor_video_kwargs : kwargs to pass to FFmpegWriter for monitor
    write_monitor_ffmpeg_stderr_to_screen : whether to display
        output from ffmpeg writing instance
    h5_filename : hdf5 file to stitch whiskers information into
    frame_func : function to apply to each frame
        If 'invert', will apply 255 - frame
    n_trace_processes : number of simultaneous trace processes
    expectedrows : how to set up hdf5 file
    verbose : verbose
    skip_stitch : skip the stitching phase

    Returns: dict
        trace_pool_results : result of each call to trace
        monitor_ff_stderr, monitor_ff_stdout : results from monitor
            video ffmpeg instance
    """
    ## Set up kwargs
    if monitor_video_kwargs is None:
        monitor_video_kwargs = {}

    if frame_func == 'invert':
        frame_func = lambda frame: 255 - frame

    # Check commands
    WhiskiWrap.utils.probe_needed_commands()

    ## Initialize readers and writers
    if verbose:
        print("initalizing readers and writers")
    # Tiff writer
    ctw = WhiskiWrap.ChunkedTiffWriter(tiffs_to_trace_directory,
        chunk_size=chunk_size, chunk_name_pattern=chunk_name_pattern)

    # FFmpeg writer is initalized after first frame
    ffw = None

    # Setup the result file
    if not skip_stitch:
        setup_hdf5(h5_filename, expectedrows, measure=True)

    # Copy the parameters files
    copy_parameters_files(tiffs_to_trace_directory, sensitive=sensitive)

    ## Set up the worker pool
    # Pool of trace workers
    trace_pool = multiprocessing.Pool(n_trace_processes)

    # Keep track of results
    trace_pool_results = []
    deleted_tiffs = []
    def log_result(result):
        trace_pool_results.append(result)

    ## Iterate over chunks
    out_of_frames = False
    nframe = 0

    # Init the iterator outside of the loop so that it persists
    iter_obj = input_reader.iter_frames()

    while not out_of_frames:
        # Get a chunk of frames
        if verbose:
            print("loading chunk of frames starting with", nframe)
        chunk_of_frames = []
        for frame in iter_obj:
            if frame_func is not None:
                frame = frame_func(frame)
            chunk_of_frames.append(frame)
            nframe = nframe + 1
            if stop_after_frame is not None and nframe >= stop_after_frame:
                break
            if len(chunk_of_frames) == chunk_size:
                break

        # Check if we ran out
        if len(chunk_of_frames) != chunk_size:
            out_of_frames = True

        ## Write tiffs
        # We do this synchronously to ensure that it happens before
        # the trace starts
        for frame in chunk_of_frames:
            ctw.write(frame)

        # Make sure the chunk was written, in case this is the last one
        # and we didn't reach chunk_size yet
        if len(chunk_of_frames) != chunk_size:
            ctw._write_chunk()
        assert ctw.count_unwritten_frames() == 0

        # Figure out which tiff file was just generated
        tif_filename = ctw.chunknames_written[-1]

        ## Start trace
        trace_pool.apply_async(trace_and_measure_chunk, args=(tif_filename, delete_tiffs, face),
            callback=log_result)

        ## Determine whether we can delete any tiffs
        #~ if delete_tiffs:
            #~ tiffs_to_delete = [
                #~ tpres['video_filename'] for tpres in trace_pool_results
                #~ if tpres['video_filename'] not in deleted_tiffs]
            #~ for filename in tiffs_to_delete:
                #~ if verbose:
                    #~ print "deleting", filename
                #~ os.remove(filename)

        ## Start monitor encode
        # This is also synchronous, otherwise the input buffer might fill up
        if monitor_video is not None:
            if ffw is None:
                ffw = WhiskiWrap.FFmpegWriter(monitor_video,
                    frame_width=frame.shape[1], frame_height=frame.shape[0],
                    write_stderr_to_screen=write_monitor_ffmpeg_stderr_to_screen,
                    **monitor_video_kwargs)
            for frame in chunk_of_frames:
                ffw.write(frame)

        ## Determine if we should pause
        while len(ctw.chunknames_written) > len(trace_pool_results) + 2 * n_trace_processes:
            print("waiting for tracing to catch up")
            time.sleep(10)

    ## Wait for trace to complete
    if verbose:
        print("done with reading and writing, just waiting for tracing")
    # Tell it no more jobs, so close when done
    trace_pool.close()

    # Wait for everything to finish
    trace_pool.join()

    ## Error check the tifs that were processed
    # Get the tifs we wrote, and the tifs we trace
    written_chunks = sorted(ctw.chunknames_written)
    traced_filenames = sorted([
        res['video_filename'] for res in trace_pool_results])

    # Check that they are the same
    if not np.all(np.array(written_chunks) == np.array(traced_filenames)):
        raise ValueError("not all chunks were traced")

    ## Extract the chunk numbers from the filenames
    # The tiffs have been written, figure out which they are
    split_traced_filenames = [os.path.split(fn)[1] for fn in traced_filenames]
    # tif_file_number_strings = wwutils.misc.apply_and_filter_by_regex(
    #     '^chunk(\d+).tif$', split_traced_filenames, sort=False)
    
    # Replace the format specifier with (\d+) in the pattern
    mod_chunk_name_pattern = re.sub(r'%\d+d', r'(\\d+)', chunk_name_pattern)
    mod_chunk_name_pattern = '^' + mod_chunk_name_pattern + '$'
    tif_file_number_strings = wwutils.misc.apply_and_filter_by_regex(
        mod_chunk_name_pattern, split_traced_filenames, sort=False)

    # Replace the format specifier with %s in the pattern
    tif_chunk_name_pattern = re.sub(r'%\d+d', r'%s', chunk_name_pattern)
    tif_full_filenames = [
        os.path.join(tiffs_to_trace_directory, tif_chunk_name_pattern % fns)
        for fns in tif_file_number_strings]
    
    tif_file_numbers = list(map(int, tif_file_number_strings))
    tif_ordering = np.argsort(tif_file_numbers)
    tif_sorted_filenames = np.array(tif_full_filenames)[
        tif_ordering]
    tif_sorted_file_numbers = np.array(tif_file_numbers)[
        tif_ordering]

    # stitch
    if not skip_stitch:
        print("Stitching")
        zobj = list(zip(tif_sorted_file_numbers, tif_sorted_filenames))
        for chunk_start, chunk_name in zobj:
            # Append each chunk to the hdf5 file
            fn = WhiskiWrap.utils.FileNamer.from_tiff_stack(chunk_name)
            append_whiskers_to_hdf5(
                whisk_filename=fn.whiskers,
		        measurements_filename = fn.measurements,
                h5_filename=h5_filename,
                chunk_start=chunk_start)

    # Finalize writers
    ctw.close()
    if ffw is not None:
        ff_stdout, ff_stderr = ffw.close()
    else:
        ff_stdout, ff_stderr = None, None

    # Also write timestamps as numpy file
    if hasattr(input_reader, 'timestamps') and timestamps_filename is not None:
        timestamps = np.concatenate(input_reader.timestamps)
        assert len(timestamps) >= ctw.frames_written
        np.save(timestamps_filename, timestamps[:ctw.frames_written])

    return {'trace_pool_results': trace_pool_results,
        'monitor_ff_stdout': ff_stdout,
        'monitor_ff_stderr': ff_stderr,
        'tif_sorted_file_numbers': tif_sorted_file_numbers,
        'tif_sorted_filenames': tif_sorted_filenames,
        }

def interleaved_split_trace_and_measure(input_reader, tiffs_to_trace_directory,
    sensitive=False,
    chunk_size=200, chunk_name_pattern='chunk%08d.tif',
    stop_after_frame=None, delete_tiffs=True,
    timestamps_filename=None, monitor_video=None,
    monitor_video_kwargs=None, write_monitor_ffmpeg_stderr_to_screen=False,
    h5_filename=None, frame_func=None,
    n_trace_processes=4, expectedrows=1000000,
    verbose=True, skip_stitch=False, face='right', classify=None, 
    summary_only=False, skip_existing=False
    ):
    """Read, write, and trace each chunk, one at a time.

    This is an extension of interleaved_read_trace_and_measure for bilateral whisker tracking.
    The function needs to be called twice, once for each side of the face. 

    input_reader : Typically a PFReader or FFmpegReader
    tiffs_to_trace_directory : Location to write the tiffs
    sensitive: if False, use default. If True, lower MIN_SIGNAL
    chunk_size : frames per chunk
    chunk_name_pattern : how to name them
    stop_after_frame : break early, for debugging
    delete_tiffs : whether to delete tiffs after done tracing
    timestamps_filename : Where to store the timestamps
        Only vallid for PFReader input_reader
    monitor_video : filename for a monitor video
        If None, no monitor video will be written
    monitor_video_kwargs : kwargs to pass to FFmpegWriter for monitor
    write_monitor_ffmpeg_stderr_to_screen : whether to display
        output from ffmpeg writing instance
    h5_filename : hdf5 file to stitch whiskers information into
    frame_func : function to apply to each frame
        If 'invert', will apply 255 - frame
    n_trace_processes : number of simultaneous trace processes
    expectedrows : how to set up hdf5 file
    verbose : verbose
    skip_stitch : skip the stitching phase
    face : 'left','right','top','bottom'
    classify : if not None, classify whiskers.

    Returns: dict
        trace_pool_results : result of each call to trace
        monitor_ff_stderr, monitor_ff_stdout : results from monitor
            video ffmpeg instance
    """
    ## Set up kwargs
    if monitor_video_kwargs is None:
        monitor_video_kwargs = {}

    if frame_func == 'invert':
        frame_func = lambda frame: 255 - frame

    if frame_func == 'fliph':
        frame_func = lambda frame: np.fliplr(frame)
    elif frame_func == 'flipv':
        frame_func = lambda frame: np.flipud(frame)

    if frame_func == 'crop':
        crop_coord = input_reader.crop
        # crop_coord format is width:height:x:y
        frame_func = lambda frame: frame[crop_coord[3]:crop_coord[3] + crop_coord[1], crop_coord[2]:crop_coord[2] + crop_coord[0]]
        # reset crop field to None
        input_reader.crop = None

    # Check commands
    WhiskiWrap.utils.probe_needed_commands()

    ## Initialize readers and writers
    if verbose:
        print("Initalizing readers and writers")

    # Tiff writer
    ctw = WhiskiWrap.ChunkedTiffWriter(tiffs_to_trace_directory,
        chunk_size=chunk_size, chunk_name_pattern=chunk_name_pattern)

    # FFmpeg writer is initalized after first frame
    ffw = None

    # Setup the result file
    if not skip_stitch:
        setup_hdf5(h5_filename, expectedrows, measure=True)

    # Copy the parameters files
    copy_parameters_files(tiffs_to_trace_directory, sensitive=sensitive)

    trace_pool_results = []

    if skip_existing:
    # Check if .whiskers or .measurements files already exist
        existing_files = [f for f in os.listdir(tiffs_to_trace_directory) if f.endswith('.whiskers') or f.endswith('.measurements')]
        if len(existing_files) > 0:
            print("Existing files found, skipping to HDF5 stitching")
            # create sorted_file_numbers, sorted_filenames
            # Keep only the .whiskers files in the list
            split_traced_filenames = [os.path.split(fn)[1] for fn in existing_files]
            split_traced_filenames = [fn for fn in split_traced_filenames if fn.endswith('.whiskers')]
            full_filenames = [os.path.join(tiffs_to_trace_directory, fn) for fn in split_traced_filenames]
            # Get the file numbers
            chunk_name_pattern = r'_(\d+)\.(whiskers)$'
            file_number_strings = [re.search(chunk_name_pattern, fn).group(1) 
                                   for fn in split_traced_filenames if re.search(chunk_name_pattern, fn)]
            # Sort the filenames by the file numbers
            file_numbers = list(map(int, file_number_strings))
            ordering = np.argsort(file_numbers)
            sorted_filenames = np.array(full_filenames)[ordering]
            sorted_file_numbers = np.array(file_numbers)[ordering]

        else:

            ## Set up the worker pool
            # Pool of trace workers
            trace_pool = multiprocessing.Pool(n_trace_processes)

            # Keep track of results
            deleted_tiffs = []
            def log_result(result):
                trace_pool_results.append(result)

            ## Iterate over chunks
            out_of_frames = False
            nframe = 0

            # Init the iterator outside of the loop so that it persists
            iter_obj = input_reader.iter_frames()

            while not out_of_frames:
                # Get a chunk of frames
                if verbose:
                    print("Loading chunk of frames starting with", nframe)
                chunk_of_frames = []
                for frame in iter_obj:
                    if frame_func is not None:
                        frame = frame_func(frame)
                    chunk_of_frames.append(frame)
                    nframe = nframe + 1
                    if stop_after_frame is not None and nframe >= stop_after_frame:
                        break
                    if len(chunk_of_frames) == chunk_size:
                        break

                # Check if we ran out
                if len(chunk_of_frames) != chunk_size:
                    out_of_frames = True

                ## Write tiffs
                # We do this synchronously to ensure that it happens before
                # the trace starts
                for frame in chunk_of_frames:
                    ctw.write(frame)

                # Make sure the chunk was written, in case this is the last one
                # and we didn't reach chunk_size yet
                if len(chunk_of_frames) != chunk_size:
                    ctw._write_chunk()
                assert ctw.count_unwritten_frames() == 0

                # Figure out which tiff file was just generated
                tif_filename = ctw.chunknames_written[-1]

                ## Start trace
                trace_pool.apply_async(trace_and_measure_chunk, args=(tif_filename, delete_tiffs, face, classify),
                    callback=log_result)

                ## Determine whether we can delete any tiffs
                #~ if delete_tiffs:
                    #~ tiffs_to_delete = [
                        #~ tpres['video_filename'] for tpres in trace_pool_results
                        #~ if tpres['video_filename'] not in deleted_tiffs]
                    #~ for filename in tiffs_to_delete:
                        #~ if verbose:
                            #~ print "deleting", filename
                        #~ os.remove(filename)

                ## Start monitor encode
                # This is also synchronous, otherwise the input buffer might fill up
                if monitor_video is not None:
                    if ffw is None:
                        ffw = WhiskiWrap.FFmpegWriter(monitor_video,
                            frame_width=frame.shape[1], frame_height=frame.shape[0],
                            write_stderr_to_screen=write_monitor_ffmpeg_stderr_to_screen,
                            **monitor_video_kwargs)
                    for frame in chunk_of_frames:
                        ffw.write(frame)

                ## Determine if we should pause
                # if len(ctw.chunknames_written) > len(trace_pool_results) + 2 * n_trace_processes:
                #     print("Waiting for tracing to catch up")
                #     #initialize time counter
                #     wait_t0 = time.time()
                #     while len(ctw.chunknames_written) > len(trace_pool_results) + 2 * n_trace_processes:
                #         # print dot every second
                #         if time.time() - wait_t0 > 1:
                #             print('.', end='')
                    # print("")
                while len(ctw.chunknames_written) > len(trace_pool_results) + 2 * n_trace_processes:
                    print("waiting for tracing to catch up")
                    time.sleep(5)

            ## Wait for trace to complete
            if verbose:
                print("Done with reading and writing, just waiting for tracing")
            # Tell it no more jobs, so close when done
            trace_pool.close()

            # Wait for everything to finish
            trace_pool.join()

            ## Error check the tifs that were processed
            # Get the tifs we wrote, and the tifs we trace
            written_chunks = sorted(ctw.chunknames_written)
            traced_filenames = sorted([
                res['video_filename'] for res in trace_pool_results])

            # Check that they are the same
            if not np.all(np.array(written_chunks) == np.array(traced_filenames)):
                raise ValueError("Not all chunks were traced")

            ## Extract the chunk numbers from the filenames
            # The tiffs have been written, figure out which they are
            split_traced_filenames = [os.path.split(fn)[1] for fn in traced_filenames]
            
            # Replace the format specifier with (\d+) in the pattern
            mod_chunk_name_pattern = re.sub(r'%\d+d', r'(\\d+)', chunk_name_pattern)
            mod_chunk_name_pattern = '^' + mod_chunk_name_pattern + '$'
            tif_file_number_strings = wwutils.misc.apply_and_filter_by_regex(
                mod_chunk_name_pattern, split_traced_filenames, sort=False)

            # Replace the format specifier with %s in the pattern
            tif_chunk_name_pattern = re.sub(r'%\d+d', r'%s', chunk_name_pattern)
            tif_full_filenames = [
                os.path.join(tiffs_to_trace_directory, tif_chunk_name_pattern % fns)
                for fns in tif_file_number_strings]
            
            tif_file_numbers = list(map(int, tif_file_number_strings))
            tif_ordering = np.argsort(tif_file_numbers)
            tif_sorted_filenames = np.array(tif_full_filenames)[
                tif_ordering]
            tif_sorted_file_numbers = np.array(tif_file_numbers)[
                tif_ordering]
            
            sorted_file_numbers = tif_sorted_file_numbers
            sorted_filenames = tif_sorted_filenames

    # stitch
    if not skip_stitch:
        print("Stitching")
        zobj = list(zip(sorted_file_numbers, sorted_filenames))
        for chunk_start, chunk_name in zobj:
            # Append each chunk to the hdf5 file
            if chunk_name.endswith('.tif'):
                fn = WhiskiWrap.utils.FileNamer.from_tiff_stack(chunk_name)
            elif chunk_name.endswith('.whiskers'):
                fn = WhiskiWrap.utils.FileNamer.from_whiskers(chunk_name)
            append_whiskers_to_hdf5(
                whisk_filename=fn.whiskers,
		        measurements_filename = fn.measurements,
                h5_filename=h5_filename,
                chunk_start=chunk_start,
                summary_only=summary_only)

    # Finalize writers
    ctw.close()
    if ffw is not None:
        ff_stdout, ff_stderr = ffw.close()
    else:
        ff_stdout, ff_stderr = None, None

    # Also write timestamps as numpy file
    if hasattr(input_reader, 'timestamps') and timestamps_filename is not None:
        timestamps = np.concatenate(input_reader.timestamps)
        assert len(timestamps) >= ctw.frames_written
        np.save(timestamps_filename, timestamps[:ctw.frames_written])

    return {'trace_pool_results': trace_pool_results,
        'monitor_ff_stdout': ff_stdout,
        'monitor_ff_stderr': ff_stderr,
        'tif_sorted_file_numbers': sorted_file_numbers,
        'tif_sorted_filenames': sorted_filenames,
        }

def interleaved_reading_and_tracing(input_reader, tiffs_to_trace_directory,
    sensitive=False,
    chunk_size=200, chunk_name_pattern='chunk%08d.tif',
    stop_after_frame=None, delete_tiffs=True,
    timestamps_filename=None, monitor_video=None,
    monitor_video_kwargs=None, write_monitor_ffmpeg_stderr_to_screen=False,
    h5_filename=None, frame_func=None,
    n_trace_processes=4, expectedrows=1000000,
    verbose=True, skip_stitch=False,
    ):
    """Read, write, and trace each chunk, one at a time.

    This is an alternative to first calling:
        write_video_as_chunked_tiffs
    And then calling
        trace_chunked_tiffs

    input_reader : Typically a PFReader or FFmpegReader
    tiffs_to_trace_directory : Location to write the tiffs
    sensitive: if False, use default. If True, lower MIN_SIGNAL
    chunk_size : frames per chunk
    chunk_name_pattern : how to name them
    stop_after_frame : break early, for debugging
    delete_tiffs : whether to delete tiffs after done tracing
    timestamps_filename : Where to store the timestamps
        Only vallid for PFReader input_reader
    monitor_video : filename for a monitor video
        If None, no monitor video will be written
    monitor_video_kwargs : kwargs to pass to FFmpegWriter for monitor
    write_monitor_ffmpeg_stderr_to_screen : whether to display
        output from ffmpeg writing instance
    h5_filename : hdf5 file to stitch whiskers information into
    frame_func : function to apply to each frame
        If 'invert', will apply 255 - frame
    n_trace_processes : number of simultaneous trace processes
    expectedrows : how to set up hdf5 file
    verbose : verbose
    skip_stitch : skip the stitching phase

    Returns: dict
        trace_pool_results : result of each call to trace
        monitor_ff_stderr, monitor_ff_stdout : results from monitor
            video ffmpeg instance
    """
    ## Set up kwargs
    if monitor_video_kwargs is None:
        monitor_video_kwargs = {}

    if frame_func == 'invert':
        frame_func = lambda frame: 255 - frame

    # Check commands
    WhiskiWrap.utils.probe_needed_commands()

    ## Initialize readers and writers
    if verbose:
        print("initalizing readers and writers")
    # Tiff writer
    ctw = WhiskiWrap.ChunkedTiffWriter(tiffs_to_trace_directory,
        chunk_size=chunk_size, chunk_name_pattern=chunk_name_pattern)

    # FFmpeg writer is initalized after first frame
    ffw = None

    # Setup the result file
    if not skip_stitch:
        setup_hdf5(h5_filename, expectedrows)

    # Copy the parameters files
    copy_parameters_files(tiffs_to_trace_directory, sensitive=sensitive)

    ## Set up the worker pool
    # Pool of trace workers
    trace_pool = multiprocessing.Pool(n_trace_processes)

    # Keep track of results
    trace_pool_results = []
    deleted_tiffs = []
    def log_result(result):
        trace_pool_results.append(result)

    ## Iterate over chunks
    out_of_frames = False
    nframe = 0

    # Init the iterator outside of the loop so that it persists
    iter_obj = input_reader.iter_frames()

    while not out_of_frames:
        # Get a chunk of frames
        if verbose:
            print("loading chunk of frames starting with", nframe)
        chunk_of_frames = []
        for frame in iter_obj:
            if frame_func is not None:
                frame = frame_func(frame)
            chunk_of_frames.append(frame)
            nframe = nframe + 1
            if stop_after_frame is not None and nframe >= stop_after_frame:
                break
            if len(chunk_of_frames) == chunk_size:
                break

        # Check if we ran out
        if len(chunk_of_frames) != chunk_size:
            out_of_frames = True

        ## Write tiffs
        # We do this synchronously to ensure that it happens before
        # the trace starts
        for frame in chunk_of_frames:
            ctw.write(frame)

        # Make sure the chunk was written, in case this is the last one
        # and we didn't reach chunk_size yet
        if len(chunk_of_frames) != chunk_size:
            ctw._write_chunk()
        assert ctw.count_unwritten_frames() == 0

        # Figure out which tiff file was just generated
        tif_filename = ctw.chunknames_written[-1]

        ## Start trace
        trace_pool.apply_async(trace_chunk, args=(tif_filename, delete_tiffs),
            callback=log_result)

        ## Determine whether we can delete any tiffs
        #~ if delete_tiffs:
            #~ tiffs_to_delete = [
                #~ tpres['video_filename'] for tpres in trace_pool_results
                #~ if tpres['video_filename'] not in deleted_tiffs]
            #~ for filename in tiffs_to_delete:
                #~ if verbose:
                    #~ print "deleting", filename
                #~ os.remove(filename)

        ## Start monitor encode
        # This is also synchronous, otherwise the input buffer might fill up
        if monitor_video is not None:
            if ffw is None:
                ffw = WhiskiWrap.FFmpegWriter(monitor_video,
                    frame_width=frame.shape[1], frame_height=frame.shape[0],
                    write_stderr_to_screen=write_monitor_ffmpeg_stderr_to_screen,
                    **monitor_video_kwargs)
            for frame in chunk_of_frames:
                ffw.write(frame)

        ## Determine if we should pause
        while len(ctw.chunknames_written) > len(trace_pool_results) + 2 * n_trace_processes:
            print("waiting for tracing to catch up")
            time.sleep(10)

    ## Wait for trace to complete
    if verbose:
        print("done with reading and writing, just waiting for tracing")
    # Tell it no more jobs, so close when done
    trace_pool.close()

    # Wait for everything to finish
    trace_pool.join()

    ## Error check the tifs that were processed
    # Get the tifs we wrote, and the tifs we trace
    written_chunks = sorted(ctw.chunknames_written)
    traced_filenames = sorted([
        res['video_filename'] for res in trace_pool_results])

    # Check that they are the same
    if not np.all(np.array(written_chunks) == np.array(traced_filenames)):
        raise ValueError("not all chunks were traced")

    ## Extract the chunk numbers from the filenames
    # The tiffs have been written, figure out which they are
    split_traced_filenames = [os.path.split(fn)[1] for fn in traced_filenames]
    tif_file_number_strings = wwutils.misc.apply_and_filter_by_regex(
        '^chunk(\d+).tif$', split_traced_filenames, sort=False)
    tif_full_filenames = [
        os.path.join(tiffs_to_trace_directory, 'chunk%s.tif' % fns)
        for fns in tif_file_number_strings]
    tif_file_numbers = list(map(int, tif_file_number_strings))
    tif_ordering = np.argsort(tif_file_numbers)
    tif_sorted_filenames = np.array(tif_full_filenames)[
        tif_ordering]
    tif_sorted_file_numbers = np.array(tif_file_numbers)[
        tif_ordering]

    # stitch
    if not skip_stitch:
        print("Stitching")
        zobj = list(zip(tif_sorted_file_numbers, tif_sorted_filenames))
        for chunk_start, chunk_name in zobj:
            # Append each chunk to the hdf5 file
            fn = WhiskiWrap.utils.FileNamer.from_tiff_stack(chunk_name)
            append_whiskers_to_hdf5(
                whisk_filename=fn.whiskers,
                h5_filename=h5_filename,
                chunk_start=chunk_start)

    # Finalize writers
    ctw.close()
    if ffw is not None:
        ff_stdout, ff_stderr = ffw.close()
    else:
        ff_stdout, ff_stderr = None, None

    # Also write timestamps as numpy file
    if hasattr(input_reader, 'timestamps') and timestamps_filename is not None:
        timestamps = np.concatenate(input_reader.timestamps)
        assert len(timestamps) >= ctw.frames_written
        np.save(timestamps_filename, timestamps[:ctw.frames_written])

    return {'trace_pool_results': trace_pool_results,
        'monitor_ff_stdout': ff_stdout,
        'monitor_ff_stderr': ff_stderr,
        'tif_sorted_file_numbers': tif_sorted_file_numbers,
        'tif_sorted_filenames': tif_sorted_filenames,
        }

def compress_pf_to_video(input_reader, chunk_size=200, stop_after_frame=None,
    timestamps_filename=None, monitor_video=None, monitor_video_kwargs=None,
    write_monitor_ffmpeg_stderr_to_screen=False, frame_func=None, verbose=True,
    ):
    """Read modulated data and compress to video

    Adapted from interleaved_reading_and_tracing

    input_reader : typically a PFReader
    chunk_size : frames per chunk
    stop_after_frame : break early, for debugging
    timestamps_filename : Where to store the timestamps
        Only valid for PFReader input_reader
    monitor_video : filename for a monitor video
        If None, no monitor video will be written
    monitor_video_kwargs : kwargs to pass to FFmpegWriter for monitor
        If None, the default is {'qp': 15} for a high-fidelity compression
        that is still ~6x smaller than lossless.
    write_monitor_ffmpeg_stderr_to_screen : whether to display
        output from ffmpeg writing instance
    frame_func : function to apply to each frame
        If 'invert', will apply 255 - frame
    verbose : verbose

    Returns: dict
        monitor_ff_stderr, monitor_ff_stdout : results from monitor
            video ffmpeg instance
    """
    ## Set up kwargs
    if monitor_video_kwargs is None:
        monitor_video_kwargs = {'qp': 15}

    if frame_func == 'invert':
        frame_func = lambda frame: 255 - frame

    ## Initialize readers and writers
    if verbose:
        print("initalizing readers and writers")

    # FFmpeg writer is initalized after first frame
    ffw = None

    ## Iterate over chunks
    out_of_frames = False
    nframe = 0
    nframes_written = 0

    # Init the iterator outside of the loop so that it persists
    iter_obj = input_reader.iter_frames()

    while not out_of_frames:
        # Get a chunk of frames
        if verbose:
            print("loading chunk of frames starting with", nframe)
        chunk_of_frames = []
        for frame in iter_obj:
            if frame_func is not None:
                frame = frame_func(frame)
            chunk_of_frames.append(frame)
            nframe = nframe + 1
            if stop_after_frame is not None and nframe >= stop_after_frame:
                break
            if len(chunk_of_frames) == chunk_size:
                break

        # Check if we ran out
        if len(chunk_of_frames) != chunk_size:
            out_of_frames = True

        ## Start monitor encode
        # This is also synchronous, otherwise the input buffer might fill up
        if monitor_video is not None:
            if ffw is None:
                ffw = WhiskiWrap.FFmpegWriter(monitor_video,
                    frame_width=frame.shape[1], frame_height=frame.shape[0],
                    write_stderr_to_screen=write_monitor_ffmpeg_stderr_to_screen,
                    **monitor_video_kwargs)
            for frame in chunk_of_frames:
                ffw.write(frame)
                nframes_written = nframes_written + 1

    # Finalize writers
    if ffw is not None:
        ff_stdout, ff_stderr = ffw.close()
    else:
        ff_stdout, ff_stderr = None, None

    # Also write timestamps as numpy file
    if hasattr(input_reader, 'timestamps') and timestamps_filename is not None:
        timestamps = np.concatenate(input_reader.timestamps)

        # These assertions only make sense if we wrote the whole file
        if stop_after_frame is None:
            assert len(timestamps) == nframes_written
            assert nframes_written == nframe

        # Save timestamps
        np.save(timestamps_filename, timestamps)

    return {
        'monitor_ff_stdout': ff_stdout,
        'monitor_ff_stderr': ff_stderr,
    }

def measure_chunk_star(args):
    return measure_chunk(*args)

def read_whiskers_hdf5_summary(filename):
    """Reads and returns the `summary` table in an HDF5 file"""
    with tables.open_file(filename) as fi:
        # Check whether the summary table exists, or updated_summary
        if '/summary' in fi:
            summary = pandas.DataFrame.from_records(fi.root.summary.read())
        elif '/updated_summary' in fi:
            try:
                # Assuming a group
                summary = pandas.DataFrame.from_records(fi.root.updated_summary.read())
            except:
                # Then assuming a table
                table_node = fi.get_node('/updated_summary/table')
                summary = pd.DataFrame.from_records(table_node.read())
        else:
            raise ValueError("no summary table found")

    return summary

def read_whiskers_measurements(filenames):
    # Loads all the whiskers/measurements files and returns an aggregated dataframe

    # Check that it is a list
    if not isinstance(filenames, list):
        filenames = [filenames]

    # Load the measurements
    all_measurements = []
    for filename in filenames:
        #  strip the extension and replace with .whiskers
        if not filename.endswith('.whiskers'):
            filename = filename.replace(filename.split('.')[-1], 'whiskers')
        # get the whisker data
        measurements = WhiskiWrap.read_whisker_data(filename, 'df')
        # append the dataframe to the list
        all_measurements.append(measurements)

    # Concatenate the dataframes
    all_measurements = pd.concat(all_measurements)

    return all_measurements

class PFReader:
    """Reads photonfocus modulated data stored in matlab files"""
    def __init__(self, input_directory, n_threads=4, verbose=True,
        error_on_unsorted_filetimes=True):
        """Initialize a new reader.

        input_directory : where the mat files are
            They are assumed to have a format like img10.mat, etc.
            They should contain variables called 'img' (a 4d array of
            modulated frames) and 't' (timestamps of each frame).
        n_threads : sent to pfDoubleRate_SetNrOfThreads

        error_on_unsorted_filetimes : bool
            Whether to raise an error if the modification times of the
            matfiles are not in sorted order, which typically happens if
            something has gone wrong (but could just be that the file times
            weren't preserved)
        """
        self.input_directory = input_directory
        self.verbose = verbose

        ## Load the libraries
        # boost_thread needs boost_system
        # I think it used to be able to find boost_system without this line
        libboost_system = ctypes.cdll.LoadLibrary(
            '/usr/local/lib/libboost_system.so.1.50.0')

        # Load boost_thread
        libboost_thread = ctypes.cdll.LoadLibrary(
            '/usr/local/lib/libboost_thread.so')

        # Load the pf_lib (which requires boost_thread)
        self.pf_lib = ctypes.cdll.LoadLibrary(LIB_DOUBLERATE)
        self.demod_func = self.pf_lib['pfDoubleRate_DeModulateImage']

        # Set the number of threads
        self.pf_lib['pfDoubleRate_SetNrOfThreads'](n_threads)

        # Find all the imgN.mat files in the input directory
        self.matfile_number_strings = wwutils.misc.apply_and_filter_by_regex(
            '^img(\d+)\.mat$', os.listdir(self.input_directory), sort=False)
        self.matfile_names = [
            os.path.join(self.input_directory, 'img%s.mat' % fns)
            for fns in self.matfile_number_strings]
        self.matfile_numbers = list(map(int, self.matfile_number_strings))
        self.matfile_ordering = np.argsort(self.matfile_numbers)

        # Sort the names and numbers
        self.sorted_matfile_names = np.array(self.matfile_names)[
            self.matfile_ordering]
        self.sorted_matfile_numbers = np.array(self.matfile_numbers)[
            self.matfile_ordering]

        # Error check the file times
        filetimes = np.array([
            wwutils.misc.get_file_time(filename)
            for filename in self.sorted_matfile_names])
        if (np.diff(filetimes) < 0).any():
            if error_on_unsorted_filetimes:
                raise IOError("unsorted matfiles")
            else:
                print("warning: unsorted matfiles")

        # Create variable to store timestamps
        self.timestamps = []
        self.n_frames_read = 0

        # Set these values once they are read from the file
        self.frame_height = None
        self.frame_width = None

    def iter_frames(self):
        """Yields frames as they are read and demodulated.

        Iterates through the matfiles in order, demodulates each frame,
        and yields them one at a time.

        The chunk of timestamps from each matfile is appended to the list
        self.timestamps, so that self.timestamps is a list of arrays. There
        will be more timestamps than read frames until the end of the chunk.

        Also sets self.frame_height and self.frame_width and checks that
        they are consistent over the session.
        """
        # Iterate through matfiles and load
        for matfile_name in self.sorted_matfile_names:
            if self.verbose:
                print(("loading %s" % matfile_name))

            # Load the raw data
            # This is the slowest step
            matfile_load = scipy.io.loadmat(matfile_name)
            matfile_t = matfile_load['t'].flatten()
            matfile_modulated_data = matfile_load['img'].squeeze()
            assert matfile_modulated_data.ndim == 3 # height, width, time

            # Append the timestamps
            self.timestamps.append(matfile_t)

            # Extract shape
            n_frames = len(matfile_t)
            assert matfile_modulated_data.shape[-1] == n_frames
            modulated_frame_width = matfile_modulated_data.shape[1]
            frame_height = matfile_modulated_data.shape[0]

            if self.verbose:
                print(("loaded %d modulated frames @ %dx%d" % (n_frames,
                    modulated_frame_width, frame_height)))

            # Find the demodulated width
            # This can be done by pfDoubleRate_GetDeModulatedWidth
            # but I don't understand what it's doing.
            if modulated_frame_width == 416:
                demodulated_frame_width = 800
            elif modulated_frame_width == 332:
                demodulated_frame_width = 640
            else:
                raise ValueError("unknown modulated width: %d" %
                    modulated_frame_width)

            # Store the frame sizes as necessary
            if self.frame_width is None:
                self.frame_width = demodulated_frame_width
            if self.frame_width != demodulated_frame_width:
                raise ValueError("inconsistent frame widths")
            if self.frame_height is None:
                self.frame_height = frame_height
            if self.frame_height != frame_height:
                raise ValueError("inconsistent frame heights")

            # Create a buffer for the result of each frame
            demodulated_frame_buffer = ctypes.create_string_buffer(
                frame_height * demodulated_frame_width)

            # Iterate over frames
            for n_frame in range(n_frames):
                if self.verbose and np.mod(n_frame, 200) == 0:
                    print(("iterator has reached frame %d" % n_frame))

                # Convert image to char array
                # Ideally we would use a buffer here instead of a copy in order
                # to save time. But Matlab data comes back in Fortran order
                # instead of C order, so this is not possible.
                frame_charry = ctypes.c_char_p(
                    matfile_modulated_data[:, :, n_frame].tobytes())
                #~ frame_charry = ctypes.c_char_p(
                    #~ bytes(matfile_modulated_data[:, :, n_frame].data))

                # Run
                self.demod_func(demodulated_frame_buffer, frame_charry,
                    demodulated_frame_width, frame_height,
                    modulated_frame_width)

                # Extract the result from the buffer
                demodulated_frame = np.fromstring(demodulated_frame_buffer,
                    dtype=np.uint8).reshape(
                    frame_height, demodulated_frame_width)

                self.n_frames_read = self.n_frames_read + 1

                yield demodulated_frame

        if self.verbose:
            print("iterator is empty")

    def close(self):
        """Currently does nothing"""
        pass

    def isclosed(self):
        return True

class ChunkedTiffWriter:
    """Writes frames to a series of tiff stacks"""
    def __init__(self, output_directory, chunk_size=200, chunk_name_pattern='chunk%08d.tif', compress=False):
        """Initialize a new chunked tiff writer.

        output_directory : where to write the chunks
        chunk_size : frames per chunk
        chunk_name_pattern : how to name the chunk, using the number of
            the first frame in it
        """
        self.output_directory = output_directory
        self.chunk_size = chunk_size
        self.chunk_name_pattern = chunk_name_pattern
        self.compress = compress

        # Initialize counters so we know what frame and chunk we're on
        self.frames_written = 0
        self.frame_buffer = []
        self.chunknames_written = []

    def write(self, frame):
        """Buffered write frame to tiff stacks"""
        # Append to buffer
        self.frame_buffer.append(frame)

        # Write chunk if buffer is full
        if len(self.frame_buffer) == self.chunk_size:
            self._write_chunk()

    def _write_chunk(self):
        if len(self.frame_buffer) != 0:
            # Form the chunk
            chunk = np.array(self.frame_buffer)

            # Name it
            chunkname = os.path.join(self.output_directory,
                self.chunk_name_pattern % self.frames_written)

            # Write it
            if self.compress:
                tifffile.imsave(chunkname, chunk, compression='lzw')
            else:
                tifffile.imsave(chunkname, chunk, compression=False)

            # Update the counter
            self.frames_written += len(self.frame_buffer)
            self.frame_buffer = []

            # Update the list of written chunks
            self.chunknames_written.append(chunkname)

    def count_unwritten_frames(self):
        """Returns the number of buffered, unwritten frames"""
        return len(self.frame_buffer)

    def close(self):
        """Finish writing any final unfinished chunk"""
        self._write_chunk()

class FFmpegReader:
    """Reads frames from a video file using ffmpeg process"""
    def __init__(self, input_filename, pix_fmt='gray', bufsize=10**9,
        duration=None, start_frame_time=None, start_frame_number=None, crop=None,
        write_stderr_to_screen=False, vsync='drop'):
        """Initialize a new reader

        input_filename : name of file
        pix_fmt : used to format the raw data coming from ffmpeg into
            a numpy array
        bufsize : probably not necessary because we read one frame at a time
        duration : duration of video to read (-t parameter)
        start_frame_time, start_frame_number : -ss parameter
            Parsed using wwutils.video.ffmpeg_frame_string
        crop : crop the video using the ffmpeg crop filter. Format is width:height:x:y
        write_stderr_to_screen : if True, writes to screen, otherwise to
            /dev/null
        """
        self.input_filename = input_filename

        # Get params
        self.frame_width, self.frame_height, self.frame_rate = \
            wwutils.video.get_video_params(input_filename, crop=crop)

        # Set up pix_fmt
        if pix_fmt == 'gray':
            self.bytes_per_pixel = 1
        elif pix_fmt == 'rgb24':
            self.bytes_per_pixel = 3
        else:
            raise ValueError("can't handle pix_fmt:", pix_fmt)
        self.read_size_per_frame = self.bytes_per_pixel * \
            self.frame_width * self.frame_height

        # Create the command
        command = ['ffmpeg']

        # Add ss string
        if start_frame_time is not None or start_frame_number is not None:
            ss_string = wwutils.video.ffmpeg_frame_string(input_filename,
                frame_time=start_frame_time, frame_number=start_frame_number)
            command += [
                '-ss', ss_string]

        # Add input file
        command += [
            '-i', input_filename,
            '-vsync', vsync,
            '-f', 'image2pipe',
            '-pix_fmt', pix_fmt]
        
        # Add crop string. The crop argument is a list
        if crop is not None:
            self.crop = crop
        else:
            self.crop = None
        #     command += [
        #         '-vf', 'crop=%d:%d:%d:%d' % tuple(crop)]

        # Add duration string
        if duration is not None:
            command += [
                '-t', str(duration),]

        # Add vcodec for pipe
        command += [
            '-vcodec', 'rawvideo', '-']

        # To store result
        self.n_frames_read = 0

        # stderr
        if write_stderr_to_screen:
            stderr = None
        else:
            stderr = open(os.devnull, 'w')

        # Init the pipe
        # We set stderr to null so it doesn't fill up screen or buffers
        # And we set stdin to PIPE to keep it from breaking our STDIN
        self.ffmpeg_proc = subprocess.Popen(command,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE, stderr=stderr,
            bufsize=bufsize)

    def iter_frames(self):
        """Yields one frame at a time

        When done: terminates ffmpeg process, and stores any remaining
        results in self.leftover_bytes and self.stdout and self.stderr

        It might be worth writing this as a chunked reader if this is too
        slow. Also we need to be able to seek through the file.
        """
        # Read this_chunk, or as much as we can
        while(True):
            raw_image = self.ffmpeg_proc.stdout.read(self.read_size_per_frame)

            # if self.crop is not None:
                # import matplotlib.pyplot as plt
                # # plot raw image, full and cropped (cropp coordinates format is width:height:x:y)
                # raw_image_full = np.frombuffer(raw_image, dtype='uint8').reshape((self.frame_height, self.frame_width, self.bytes_per_pixel))
                # raw_image_cropped = raw_image_full[self.crop[3]:self.crop[3]+self.crop[1], self.crop[2]:self.crop[2]+self.crop[0], :]
                # plt.figure()
                # plt.subplot(1,2,1)
                # plt.imshow(raw_image_full)
                # plt.subplot(1,2,2)
                # plt.imshow(raw_image_cropped)
                # plt.show()

            # check if we ran out of frames
            if len(raw_image) != self.read_size_per_frame:
                self.leftover_bytes = raw_image
                self.close()
                return

            # Convert to array, flatten and squeeze (and crop if required, although it is slightly faster to crop in the calling function with frame_func)
            if self.crop is None:
                frame = np.frombuffer(raw_image, dtype='uint8').reshape((self.frame_height, self.frame_width, self.bytes_per_pixel)).squeeze()
            else:
                frame = np.frombuffer(raw_image, dtype='uint8').reshape((self.frame_height, self.frame_width, self.bytes_per_pixel)).squeeze()[self.crop[3]:self.crop[3]+self.crop[1], self.crop[2]:self.crop[2]+self.crop[0]]

            # Update
            self.n_frames_read = self.n_frames_read + 1

            # Yield
            yield frame

    def close(self):
        """Closes the process"""
        # Need to terminate in case there is more data but we don't
        # care about it
        # But if it's already terminated, don't try to terminate again
        if self.ffmpeg_proc.returncode is None:
            self.ffmpeg_proc.terminate()

            # Extract the leftover bits
            self.stdout, self.stderr = self.ffmpeg_proc.communicate()

        return self.ffmpeg_proc.returncode

    def isclosed(self):
        if hasattr(self.ffmpeg_proc, 'returncode'):
            return self.ffmpeg_proc.returncode is not None
        else:
            # Never even ran? I guess this counts as closed.
            return True

class FFmpegWriter:
    """Writes frames to an ffmpeg compression process"""
    def __init__(self, output_filename, frame_width, frame_height,
        output_fps=30, vcodec='libx264', qp=15, preset='medium',
        input_pix_fmt='gray', output_pix_fmt='yuv420p',
        write_stderr_to_screen=False):
        """Initialize the ffmpeg writer

        output_filename : name of output file
        frame_width, frame_height : Used to inform ffmpeg how to interpret
            the data coming in the stdin pipe
        output_fps : frame rate
        input_pix_fmt : Tell ffmpeg how to interpret the raw data on the pipe
            This should match the output generated by frame.tostring()
        output_pix_fmt : pix_fmt of the output
        crf : quality. 0 means lossless
        preset : speed/compression tradeoff
        write_stderr_to_screen :
            If True, writes ffmpeg's updates to screen
            If False, writes to /dev/null

        With old versions of ffmpeg (jon-severinsson) I was not able to get
        truly lossless encoding with libx264. It was clamping the luminances to
        16..235. Some weird YUV conversion?
        '-vf', 'scale=in_range=full:out_range=full' seems to help with this
        In any case it works with new ffmpeg. Also the codec ffv1 will work
        but is slightly larger filesize.
        """
        # Open an ffmpeg process
        cmdstring = ('ffmpeg',
            '-y', '-r', '%d' % output_fps,
            '-s', '%dx%d' % (frame_width, frame_height), # size of image string
            '-pix_fmt', input_pix_fmt,
            '-f', 'rawvideo',  '-i', '-', # raw video from the pipe
            '-pix_fmt', output_pix_fmt,
            '-vcodec', vcodec,
            '-qp', str(qp),
            '-preset', preset,
            output_filename) # output encoding

        if write_stderr_to_screen:
            self.ffmpeg_proc = subprocess.Popen(cmdstring, stdin=subprocess.PIPE,
                stdout=subprocess.PIPE)
        else:
            self.ffmpeg_proc = subprocess.Popen(cmdstring, stdin=subprocess.PIPE,
                stdout=subprocess.PIPE, stderr=open('/dev/null', 'w'))

    def write(self, frame):
        """Write a frame to the ffmpeg process"""
        self.ffmpeg_proc.stdin.write(frame.tostring())

    def write_bytes(self, bytestring):
        self.ffmpeg_proc.stdin.write(bytestring)

    def close(self):
        """Closes the ffmpeg process and returns stdout, stderr"""
        return self.ffmpeg_proc.communicate()

